from PySide6 import QtWidgets, QtCore, QtGui

class OptionExpiryCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, text):
        super().__init__(text)
        self.setBackground(QtGui.QBrush(QtGui.QColor("orange")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("white")))
        self.setTextAlignment(QtCore.Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

class OptionMetricCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, text):
        super().__init__(text)
        self.setBackground(QtGui.QBrush(QtGui.QColor("black")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("white")))
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

class CustomDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, table):
        super().__init__()
        self.table = table

    def paint(self, painter, option, index):
        item = self.table.item(index.row(), index.column())
        if item and item.background().color() == QtGui.QColor("orange"):
            painter.fillRect(option.rect, QtGui.QColor("orange"))
            painter.save()
            painter.setFont(item.font())
            painter.setPen(item.foreground().color())
            text_rect = option.rect.adjusted(4, 0, -4, 0)
            painter.drawText(
                text_rect,
                item.textAlignment() | QtCore.Qt.AlignVCenter,
                item.text()
            )
            painter.restore()
        else:
            super().paint(painter, option, index)

def create_child_table():
    child_table = QtWidgets.QTableWidget(50, 4)
    child_table.setItemDelegate(CustomDelegate(child_table))
    child_table.setStyleSheet("""
        QTableWidget {
            background-color: black;
            border-left: 2px solid #414141;
            border-right: 2px solid #414141;
        }
        QTableWidget::item {
            border: 0px;
            padding: 4px;
        }
        QTableWidget::item:selected {
            background-color: #414141;
        }
    """)
    child_table.setShowGrid(False)
    child_table.verticalHeader().setVisible(False)
    child_table.horizontalHeader().setVisible(True)
    child_table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
    child_table.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

    for row in range(child_table.rowCount()):
        for col in range(child_table.columnCount()):
            if row == 0:
                item = OptionExpiryCellItem(f"Expiry {col+1}")
            else:
                item = OptionMetricCellItem(f"{row}.{col}")
            child_table.setItem(row, col, item)

    child_table.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                               QtWidgets.QSizePolicy.Expanding)
    return child_table

def link_scrollbars(*tables):
    for table in tables:
        table.verticalScrollBar().valueChanged.connect(
            lambda value, current=table: sync_scroll(value, current, tables)
        )

def sync_scroll(value, source, tables):
    for table in tables:
        if table is not source:
            table.verticalScrollBar().setValue(value)

app = QtWidgets.QApplication([])
window = QtWidgets.QMainWindow()

parent_table = QtWidgets.QTableWidget(1, 3)

child_table1 = create_child_table()
child_table2 = create_child_table()
child_table3 = create_child_table()

parent_table.setCellWidget(0, 0, child_table1)
parent_table.setCellWidget(0, 1, child_table2)
parent_table.setCellWidget(0, 2, child_table3)

parent_table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
parent_table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

link_scrollbars(child_table1, child_table2, child_table3)

window.setCentralWidget(parent_table)
window.resize(800, 600)
window.show()
app.exec()



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from pyvol_terminal.engines.interpolation_engines import Abstract3DInterpolator
    from pyvol_terminal.engines.surface_engines import AbstractSurfaceEngine
    from quantities.engines import MetricAxisEngine
from pydantic.dataclasses import dataclass, Field
from dataclasses import InitVar

import numpy as np
from typing import Any, List
from typing import Dict, Callable
from . import classes


class Config:
    arbitrary_types_allowed = True

@dataclass(slots=True, kw_only=True, config=Config)
class BaseContainer:
    all_price_types: List[str]
    displayed_price_types: List[str]
    instrument_names: List[str]
    metric_axis_engine: "MetricAxisEngine"
    domain: classes.Domain
    raw: Dict[str, classes.OptionChain]
    

    def update_value_by_instrument_object(self, instrument_object):
        for px_type, raw in self.raw.items():
            raw.update_all_metrics_name(instrument_object.ticker,
                                        *instrument_object.get_all_metrics_price_type(px_type))
            


@dataclass(slots=True)
class SurfaceContainer(BaseContainer):    
    surface_dataclasses: Dict[str, classes.Surface]
    scatter_dataclasses: Dict[str, classes.Points]
    x_min: float = Field(init=False, default=np.nan)
    x_max: float = Field(init=False, default=np.nan)
    y_min: float = Field(init=False, default=np.nan)
    y_max: float = Field(init=False, default=np.nan)
    z_min: float = Field(init=False, default=np.nan)
    z_max: float = Field(init=False, default=np.nan)
    valid_surface: bool = Field(init=False)
    valid_data_callback: List[Callable] = Field(init=False, default_factory=list)
    displayed_dataclasse_types: Dict[str, bool] = Field(default_factory=lambda: {"surface": True, "scatter": True})
    
    def __post_init__(self):
        self.calculate_limits()
        
    def add_valid_data_callback(self, callback):
        self.valid_data_callback.append(callback)
    
    def switch_axis(self, axis_direction, new_label):
        if new_label == self.metric_axis_engine.get_metric(axis_direction):
            return
        else:
            self.metric_axis_engine.change_function(axis_direction, new_label)
            self.cleanup()
            
    def toggle_data_dataclass(self, dataclass_type):
        self.displayed_dataclasse_types[dataclass_type]^=True
        
        dataclasses = getattr(self, f"{dataclass_type}_dataclasses")
        for dataclass in dataclasses:
            dataclass.displayed^=True
        
    def add_displayed_price_type(self, px_type):
        self.displayed_price_types.append(px_type)
        x, y, z, _ = self.metric_axis_engine.transform_values(self.raw[px_type])
        
        self.surface_dataclasses[px_type].evaluate_from_engine(x, y, z)
        self.scatter_dataclasses[px_type].evaluate_from_engine(x, y, z)
    
    def toggle_displayed_price_type(self, px_type):
        print(f"\ncontainers.toggle_displayed_price_type: {px_type}")
        if px_type in self.displayed_price_types:
            self.remove_displayed_price_type(px_type)
        else:
            self.add_displayed_price_type(px_type)
        self.calculate_limits()
    
    def remove_displayed_price_type(self, px_type):
        self.displayed_price_types.remove(px_type)

    def cleanup(self):
        for px_type in self.displayed_price_types:
            x, y, z, _ = self.metric_axis_engine.transform_values(self.raw[px_type])
            self.surface_dataclasses[px_type].evaluate_from_engine(x, y, z)
            self.scatter_dataclasses[px_type].evaluate_from_engine(x, y, z)
        self.calculate_limits()
        
    def get_dataclasses(self, px_type):
        return [self.surface_dataclasses[px_type], self.scatter_dataclasses[px_type]]
        
    def calculate_limits(self):
        if len(self.displayed_price_types) > 0:
            x_mins=[]
            x_maxs=[]
            y_mins=[]
            y_maxs=[]
            for px_type in self.displayed_price_types:
                for data_class in self.get_dataclasses(px_type):
                    if data_class.displayed:
                        x_mins.append(data_class.x_min)
                        x_maxs.append(data_class.x_max)
                        y_mins.append(data_class.y_min)
                        y_maxs.append(data_class.y_max)
                        
            if len(x_mins) > 0:
                self.x_min = np.amin(x_mins)
                self.x_max = np.amax(x_maxs)
                self.y_min = np.amin(y_mins)
                self.y_max = np.amax(y_maxs)
                
            z_mins=[]
            z_maxs=[]

            for px_type in self.displayed_price_types:
                for data_class in self.get_dataclasses(px_type):
                    if data_class.valid_values and data_class.displayed:
                        z_mins.append(data_class.z_min)
                        z_maxs.append(data_class.z_max)

            if len(z_mins) > 0:
                self.z_min = np.amin(z_mins)  
                self.z_max = np.amax(z_maxs)
                self.valid_surface = True
                for callback in self.valid_data_callback:
                    callback(True)
            else:
                self.valid_surface = False
                for callback in self.valid_data_callback:
                    callback(False)

    def get_limits(self):
        return self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max

    
@dataclass(slots=True)
class SliceContainer(BaseContainer):
    interpolator_engine: InitVar[Any]
    n_x: int
    axis_direction: str
    slice_container: Dict[str, classes.Slice]
    
    domain_idx_map: Dict[str, int] = None
    domain_vec: Dict[str, np.ndarray] = None
    domain_matrix: Dict[str, np.ndarray]  = None
    displayed_slices: List[str] = Field(init=False)  
    z_values: Dict[str, np.ndarray] = Field(init=False)
    x_metric: str = ""
    y_metric: str = ""
    
    
    def filter_domain(self):
        value_dict = {}
        for value in self.domain_vec:
            if self.axis_direction == 0:
                mask = self.domain.base_domain.expiry == value
                filtered_domain = self.domain.base_domain.expiry[mask]
            else:
                mask = self.domain.base_domain.strike == value
                filtered_domain = self.domain.base_domain.strike[mask]
            value_dict[value] = len(filtered_domain)
        top_12 = sorted(value_dict, key=value_dict.get, reverse=True)[:12]
        top_12 = np.sort(top_12)
        self.domain_vec = self.domain_vec[np.isin(self.domain_vec, top_12)]
        
    def update_value_by_instrument_object(self, instrument_object):
        for px_type, raw_dataclass in self.raw.items():
            raw_dataclass.update_all_metrics_name(instrument_object.ticker,
                                                  *instrument_object.get_all_metrics_price_type(px_type))
    
    def add_slice(self, px_type, value):
        self.update_slice(px_type, value)
        self.displayed_slices.append(value)
        
    def remove_slice(self, value):
        self.displayed_slices.remove(value)
    
    def switch_axis(self, new_axis_label, axis_direction):
        metric = self.metric_axis_engine.label_metric_map[new_axis_label]
        self.transformer = self.metric_axis_engine.get_function(metric)
        setattr(self, f"{axis_direction}_metric", metric)
        
        for px_type in self.displayed_price_types:
            self.update_slices(px_type)
    
    
    def update_slice(self, px_type, value):
        idx = self.domain_idx_map[value]
        x_old = self.domain_matrix[px_type]
        y_old = self.raw[px_type].ivol
        
        if self.x_metric == "delta":
            x_old = self.domain.x.flatten()
            y_old = self.domain.y.flatten()
            z_old = self.raw[px_type].ivol.flatten()
            x, y, z, idx_remove, idx_sort = self.transformer(self.raw[px_type],
                                                            x=x_old,
                                                            y=y_old,
                                                            z=z_old,
                                                            flatten=True)
            mask = y == value
            x, y, z = x[mask], y[mask], z[mask] 

            x_all = self.domain.x.flatten()[idx_remove][idx_sort][mask]
            nan_vec_x = np.full_like(self.domain.x_vect, np.nan)
            nan_vec_y = np.full_like(self.domain.x_vect, np.nan)
            
            mask2 = np.isin(self.domain.x_vect, x_all)
            
            nan_vec_x[mask2] = x
            nan_vec_y[mask2] = z
            
            x = nan_vec_x
            z = nan_vec_y        
        else:
            x, y, z, idx_remove, idx_sort = self.transformer(self.raw[px_type],
                                                            x=x_old,
                                                            z=y_old)
        if self.x_metric != "delta":
            if self.axis_direction == 0:
                x = x[idx]
                z = z[idx]
                self.domain_matrix[px_type][idx] = x 
                self.z_values[px_type][idx] = z
            else:
                x = x[:,idx]
                z = z[:,idx]
                self.domain_matrix[px_type][:, idx] = x
                self.z_values[px_type][:, idx] = z
        else:
            self.domain_matrix[px_type][idx] = x
            self.z_values[px_type][idx] = z

            
        if self.x_metric != "delta":
            xi = np.linspace(np.nanmin(x), np.nanmax(x), self.n_x)
        else:
            
            xi = np.linspace(0, 1, self.n_x)
            
        slice = self.slice_container[px_type][value]
        slice.update_domain(x, xi)
        slice.interpolate(x, z)
        
    def update_slices(self, px_type):
        for slice_value in self.displayed_slices:
            self.update_slice(px_type, slice_value)

    def cleanup(self):
        for px_type in self.displayed_price_types:
            self.update_slices(px_type)
            for slice_value in self.displayed_slices:
                slice = self.slice_container[px_type][slice_value]
                try:
                    if self.axis_direction == 0:
                        slice.interpolate(self.domain_matrix[px_type][self.domain_idx_map[slice_value]],
                                        self.z_values[px_type][self.domain_idx_map[slice_value]])
                    else:
                        slice.interpolate(self.domain_matrix[px_type][:, self.domain_idx_map[slice_value]],
                                        self.z_values[px_type][:, self.domain_idx_map[slice_value]])
                except:
                    pass



#%%


import QuantLib as ql
from datetime import datetime
import numpy as np
import pandas as pd
import examples.deribit.generate_synthetic_data as gen_syn_data 
import json 
from pyvol_terminal.engines import interpolation_engines
import Heston_COS_METHOD

def get_saved_data():
    df_options = pd.read_csv("df_options.csv")
    df_futures = pd.read_csv("df_futures.csv")
    df_spot = pd.read_csv("df_spot.csv")
    
    
    #expiry = df_options["expiry"].unique()[:8]
    df_options = df_options[(df_options["strike"] <= 100_000) & (df_options["strike"] >= 70_000)]
    df_options = df_options[df_options["strike"] % 5000 == 0]
    
    with open("option_underlying_ticker_map.json", "r") as file:
        option_underlying_ticker_map = json.load(file)
        
    with open("future_underlying_ticker_map.json", "r") as file:
        future_underlying_ticker_map = json.load(file)

    with open("channels.txt", "r") as file:
        channels = [line.strip() for line in file]  
        
    return channels, df_options, df_futures, df_spot, option_underlying_ticker_map, future_underlying_ticker_map



channels, df_options, df_futures, df_spot, option_underlying_ticker_map, future_underlying_ticker_map = get_saved_data()


vol_params = {"ATM_vol" : 0.65,
            "kurtosis_1" : 0.55,
            "kurtosis_2" : 0.2,
            "decay" : 0.1}


df_options, df_futures, df_spot = gen_syn_data.generate_synth_data(vol_params, df_options, df_futures, df_spot)

def create_matrices(df):
    x = df["strike"]
    y = df["expiry"]
    z = df["mid_iv"]

    instrument_names = df["instrument_name"].to_list()
    
    x_unique = np.unique(x)
    y_unique = np.unique(y)
    
    X_grid, Y_grid = np.meshgrid(x_unique, y_unique)
    Z_grid = np.full_like(X_grid, np.nan)
    name_index_map = {}
    for xi, yi, zi, name in zip(x, y, z, instrument_names):
        i = np.where(y_unique == yi)[0][0] 
        j = np.where(x_unique == xi)[0][0] 
        Z_grid[i, j] = zi
        name_index_map[name] = (i,j)
    
    missing_points = []
    
    for i in range(Z_grid.shape[0]):
        for j in range(Z_grid.shape[1]):
            if np.isnan(Z_grid[i,j]):
                missing_points.append((i,j))

    return X_grid, Y_grid, Z_grid, name_index_map , missing_points

X_grid, Y_grid, Z_grid, name_index_map, missing_points = create_matrices(df_options)

X_flat = X_grid.flatten()
Y_flat = Y_grid.flatten()
Z_flat = Z_grid.flatten()

mask = ~np.isnan(Z_flat)

X_flat, Y_flat, Z_flat = X_flat[mask], Y_flat[mask], Z_flat[mask]


interpolator = interpolation_engines.CustomBSplineInterpolator()
interpolator.fit(X_flat, Y_flat, Z_flat)

xi = X_grid.flatten()
yi = Y_grid.flatten()

data = interpolator.evaluate(xi, yi)


day_count = ql.Actual365Fixed()
calendar = ql.UnitedStates(m=1)

t_0 = 1743506630.313821
dt_0 = datetime.fromtimestamp(t_0)

calculation_date = ql.Date(dt_0.day, dt_0.month, dt_0.year)
spot = 0.5 * (df_spot['bids'].item() + df_spot['asks'].item())
ql.Settings.instance().evaluationDate = calculation_date


risk_free_rate = 0.01
dividend_rate = 0.00
dividend_yield = ql.QuoteHandle(ql.SimpleQuote(0.0))

flat_ts = ql.YieldTermStructureHandle(ql.FlatForward(calculation_date, risk_free_rate, day_count))
dividend_ts = ql.YieldTermStructureHandle(ql.FlatForward(calculation_date, dividend_rate, day_count))

strikes = np.unique(X_grid)

expiration_ts = np.unique(Y_grid)

expiration_dt = [datetime.fromtimestamp(ts) for ts in expiration_ts]    

expiration_dates = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt]

v0, kappa, theta, rho, sigma = 0.01, 0.2, 0.02, -0.75, 0.5

process = ql.HestonProcess(flat_ts, dividend_ts,
                           ql.QuoteHandle(ql.SimpleQuote(spot)),
                           v0, kappa, theta, sigma, rho)

model = ql.HestonModel(process)
engine = ql.AnalyticHestonEngine(model)
heston_helpers = []


for j, date in enumerate(expiration_dates):
    for i, s in enumerate(strikes):
        t = (date - calculation_date)
        p = ql.Period(t, ql.Days)
        sigma = data[i][j]
        helper = ql.HestonModelHelper(p, calendar, spot, s,
                                    ql.QuoteHandle(ql.SimpleQuote(sigma)),
                                    flat_ts,
                                    dividend_ts)
        helper.setPricingEngine(engine)
        heston_helpers.append(helper)
lm = ql.LevenbergMarquardt(1e-8, 1e-8, 1e-8)
model.calibrate(heston_helpers, lm, ql.EndCriteria(500, 50, 1.0e-8, 1.0e-8, 1.0e-8))
theta, kappa, sigma, rho, v0 = model.params()

print ("\ntheta = %f, kappa = %f, sigma = %f, rho = %f, v0 = %f" % (theta, kappa, sigma, rho, v0))
avg = 0.0

print ("%15s %15s %15s %20s" % (
    "Strikes", "Market Value",
     "Model Value", "Relative Error (%)"))
print ("="*70)
for i, opt in enumerate(heston_helpers):
    err = (opt.modelValue()/opt.marketValue() - 1.0)
    print ("%15.2f %15.5f %20.7f " % (
        opt.marketValue(),
        opt.modelValue(),
        100.0*(opt.modelValue()/opt.marketValue() - 1.0)))
    avg += abs(err)
avg = avg*100.0/len(heston_helpers)
print ("-"*70)
print ("Average Abs Error (%%) : %5.3f" % (avg))
# %%

N = 100
L = 20
r = 0
q = 0

expiry_years = (Y_flat - t_0) /3600 / 24/365


riskFreeTS = ql.YieldTermStructureHandle(ql.FlatForward(calculation_date, 0.05, ql.Actual365Fixed()))
dividendTS = ql.YieldTermStructureHandle(ql.FlatForward(calculation_date, 0.01, ql.Actual365Fixed()))

initialValue = ql.QuoteHandle(ql.SimpleQuote(spot))


hestonProcess = ql.HestonProcess(riskFreeTS, dividendTS, initialValue, v0, kappa, theta, sigma, rho)
hestonModel = ql.HestonModel(hestonProcess)

engine = ql.AnalyticHestonEngine(hestonModel)



#%%
flags = [1 if k > spot else 1 for k in X_flat]


expiration_dt_flat = [datetime.fromtimestamp(ts) for ts in Y_flat]    

expiration_dates_flat = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt_flat]


prices = Heston_COS_METHOD.heston_cosine_method(spot, X_flat, expiry_years, N, L, r, q, theta, v0, sigma, rho, kappa, flags)

for idx, price in enumerate(prices.flatten()):
    
    strike = X_flat[idx]
    if strike > spot:
        ql_opt_type = ql.Option.Call
    else:
        ql_opt_type = ql.Option.Put
    
    payoff = ql.PlainVanillaPayoff(ql_opt_type, strike)
    exercise = ql.EuropeanExercise(expiration_dates_flat[idx])
    europeanOption = ql.VanillaOption(payoff, exercise)

    europeanOption.setPricingEngine(engine)
    
    print("")
    print(f"mine: {price}, ql: {europeanOption.NPV()}")
#%%



volTS = ql.BlackVolTermStructureHandle(ql.BlackConstantVol(calculation_date, ql.TARGET(), 0, ql.Actual360()))
process = ql.GeneralizedBlackScholesProcess(initialValue, dividendTS, riskFreeTS, volTS)
europeanOption.impliedVolatility(europeanOption.NPV(), process)

#%%%
from examples.deribit import generate_synthetic_data
vol_params = {"ATM_vol" : 0.65,
                           "kurtosis_1" : 0.55,
                           "kurtosis_2" : 0.2,
                            "decay" : 0.1}


channels, df_options, df_futures, df_spot, option_underlying_ticker_map, future_underlying_ticker_map = generate_synthetic_data.get_saved_data()
df_options, df_futures, df_spot = generate_synthetic_data.generate_synth_data(vol_params, df_options, df_futures, df_spot)
messages_options, messages_futures, messages_spot = generate_synthetic_data.create_starting_messages(df_options, df_futures, df_spot)






from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, ClassVar

from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GLViewBox import GLViewBox
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from pyqtgraph.Qt import QT_LIB
import importlib
import pyqtgraph as pg
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DViewBox
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DViewBox, GLSurfacePlotItem, AbstractGLPlotItem, AbstractGLGraphicsItem, GLScatterPlotItem
from pyqtgraph import opengl
import sys
import cProfile
import sys
import pstats
from pyqtgraph import ButtonItem, icons
from dataclasses import dataclass, InitVar, field
from abc import ABC, abstractmethod, ABCMeta
import abc
from PySide6 import QtGui, QtCore, QtWidgets
from PySide6.QtWidgets import QGraphicsItem
from enum import Enum, EnumType, Flag, IntFlag
from pyqtgraph import Transform3D
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
import warnings
import weakref

class QABCMeta(ABCMeta, type(QtCore.QObject)):
    """Meta class for OpenGL mixin objects, combining ABC and QObject."""
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        print(f"cls: {cls}")
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)


class GLGraphicsItemMixin(GLGraphicsItem, metaclass=QABCMeta):
    def __init__(self, parentItem: 'GLGraphicsItemMixin'=None, **kwargs):
        self.__parent = None
        self.__view = None
        self.__children = list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__initialized = False
        self.__glOpts = {}
        print("GLGraphicsItemMixin")
        super().__init__(parentItem=parentItem, **kwargs)
        
    def _paintHelper(self, *args, **kwargs):
        self.childPaint()
    
    @abstractmethod
    def childPaint(self):...

class _BaseGLPlotDataItemMixin(GLGraphicsItemMixin):
    sigPlotChanged = QtCore.Signal(object) 
    
    def __init__(self, parentItem=None, **kwargs):
        print("GLPlotDataItemMixin")
        super().__init__(parentItem=parentItem, **kwargs)
        
class MeshGLPlotDataItemMixin(_BaseGLPlotDataItemMixin):
    def __init__(self, parentItem=None, **kwargs):
        print("GLPlotDataItemMixin")
        super().__init__(parentItem=parentItem, **kwargs)

        

class CustomPlotDataItemClass(opengl.GLSurfacePlotItem):
    def __init__(self, parentItem=None, **kwargs):
        print("CustomPlotDataItemClass")
        super().__init__(parentItem=parentItem, **kwargs)
        
    def childPaint(self):
        return opengl.GLSurfacePlotItem.paint(self)

class mixedPlotDataItemClass(MeshGLPlotDataItemMixin, CustomPlotDataItemClass):
    def __init__(self, x=None, y=None, z=None, colors=None, parentItem=None, **kwds):
        print("mixedPlotDataItemClass")
        super().__init__(parentItem=parentItem, x=x, y=y, z=z, colors=colors, **kwds)

    def childPaint(self):
        return opengl.GLTextItem.paint(self)

class customGraphicsItemClass(opengl.GLTextItem):
    def __init__(self, parentItem=None, **kwargs):
        super().__init__(parentItem=parentItem, **kwargs)
    

class mixedCustomGraphicsItemClass(GLGraphicsItemMixin, customGraphicsItemClass):
    def __init__(self, text="Hello", pos=None, color=(1, 1, 1, 1), font=None, parentItem=None, **kwds):
        print("Mixed Text Item")
        super().__init__(text=text, pos=pos, color=color, font=font, parentItem=parentItem, **kwds)
        
    def childPaint(self):
        return opengl.GLTextItem.paint(self)

class customGLGraphItemClass(opengl.GLGraphItem):
    def __init__(self, parentItem=None, **kwargs):
        super().__init__(parentItem=parentItem, **kwargs)

class mixedcustomGLGraphItemClass(GLGraphicsItemMixin, customGLGraphItemClass):
    def __init__(self, text="Hello", pos=None, color=(1, 1, 1, 1), font=None, parentItem=None, **kwds):
        print("Mixed customGLGraphItem Item")
        super().__init__(text=text, pos=pos, color=color, font=font, parentItem=parentItem, **kwds)


def create_pos():
    X = np.arange(-5, 5, 0.5)
    Y = np.arange(-3, 3, 0.5)
    X_mat, Y_mat = np.meshgrid(X, Y, indexing="xy")
    R = np.sqrt(X_mat**2 + Y_mat**2)

    m1 = 1
    
    Z_mat = m1 * np.sin(R) 
    return X_mat, Y_mat, Z_mat, m1

class CustomColorMap(pg.ColorMap):
    def __init__(self, colourmap_style: str):
        pos = np.linspace(0, 1, 500)          
        try:
            colourmap = pg.colormap.get(colourmap_style)
        except:
            print(f"{colourmap_style} is not in pyqtgraph, using default inferno")
            colourmap = pg.colormap.get("inferno")

        colors = colourmap.map(pos, mode='byte')        
        super().__init__(pos=pos, color=colors, mode='byte')

def create_objects(X_mat, Y_mat, Z_mat, m1) -> AbstractGLPlotItem:

    
    X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
    
    pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
    """
    scatter = GLScatterPlotItem.GLScatterPlotItem(pos=pos_scatter,
                                                  glOptions='opaque',
                                                  color=(1, 0, 1, 1)
                                                  )
    """
    scatter=None
    pos_surface = X_mat[0], Y_mat[:,1], Z_mat.T
    colormap = CustomColorMap("inferno")
    
    z_norm = (Z_mat - Z_mat.min()) / (Z_mat.max() - Z_mat.min())
    colors = colormap.map(np.linspace(0, 1, 500), mode='byte')
    # Map normalized Z to RGB colors


    
    surface = mixedPlotDataItemClass(x=pos_surface[0],
                                                  y=pos_surface[1],
                                                  z=pos_surface[2],
                                                  glOptions='opaque',
                                                  shader="shaded",
                                                  colors=colors
                                                  )
    
    print(f"surface._x: {surface._x}")
    return scatter, surface




class Window(QtWidgets.QMainWindow):
    def __init__(self, *args):
        super().__init__(*args)
        
        self.gl_view = opengl.GLViewWidget()
        """
        worldRange=[[0.,1.], [0.,1.],[0.,1.]]

        self.viewBox = GLViewBox.GLViewBox(worldRange=worldRange)
        
        self.gl_view = GLPlotWidget.GLPlotWidget(
                                            # worldRange=worldRange,
                                             #worldRange=np.array([[0.,1.]]*3)
                                             viewBox=self.viewBox
                                             )

        """
   #     self.central_layout = QtWidgets.QVBoxLayout(self)
    #    self.central_layout.addWidget(self.gl_view)
        self.setCentralWidget(self.gl_view)
        self.gl_view.opts["distance"]=5
        self.gl_view.update()
        self.setCentralWidget(self.gl_view)
        self.setGeometry(0, 0, 500, 300)
        self.show()
        self.X_mat, self.Y_mat, self.Z_mat, self.m1 = create_pos()
        self.scatter, self.surface = create_objects(self.X_mat, self.Y_mat, self.Z_mat, self.m1)    
        
        self.gl_view.addItem(self.surface)
        
        
def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
   # win.timer.start(750)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()
    
#%%
"""
from abc import ABC, abstractmethod, ABCMeta
import abc
from pyqtgraph import opengl, Transform3D


class GObj:
    def __init__(self, *args, **kwargs):
        print("GP1")


class pyqtGraphicsItem(GObj):
    def __init__(self, *args, **kwargs):
        print("GP2")
        super().__init__(*args,  **kwargs)

class GLMeta(abc.ABCMeta, type(GObj)):
    pass

class Parent1(pyqtGraphicsItem):
    def __init__(self, *args, **kwargs):
        print("Parent1")
        super().__init__(*args,  **kwargs)

    def paint(self):
        print("parent1 paint")



class GLGraphicsItemMixin(pyqtGraphicsItem, metaclass=GLMeta):
    def __init__(self, parentItem: 'GLGraphicsItemMixin'=None, **kwargs):
        self.__parent = None
        self.__view = None
        self.__children = list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__initialized = False
        self.__glOpts = {}
        print("GLGraphicsItemMixin")
        super().__init__(parentItem=parentItem, **kwargs)
        
    def _paintHelper(self, *args, **kwargs):
        self.childPaint()
    
    @abstractmethod
    def childPaint(self):...
 


class GLPlotDataItemMixin(GLGraphicsItemMixin):

    def __init__(self, parentItem=None, **kwargs):
        print("GLPlotDataItemMixin")
        super().__init__(parentItem=parentItem, **kwargs)

class CustomPlotDataItemClass(Parent1):
    def __init__(self, parentItem=None, **kwargs):
        print("CustomPlotDataItemClass")
        super().__init__(parentItem=parentItem, **kwargs)
        
class mixedPlotDataItemClass(GLPlotDataItemMixin, CustomPlotDataItemClass):
    def __init__(self, x=None, y=None, z=None, colors=None, parentItem=None, **kwds):
        print("mixedPlotDataItemClass")
        super().__init__(parentItem=parentItem, x=x, y=y, z=z, colors=colors, **kwds)

    def childPaint(self):
        return Parent1.paint(self)




mixedPlotDataItemClass()
"""


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from ...engines.interpolation_engines import Abstract3DInterpolator
    from ...data_classes.classes import VolatilityData, VolVector

from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from . import stylesheets as stylesheets, settings, main_view
from pyvol_terminal.misc_classes import PriceText
from .. import abstract_classes
from .main_view import MainView

class Interface(abstract_classes.ABCInterface):
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 tick_label_engine=None,
                 parent=None,
                 **kwargs):
        super().__init__(parent=parent)
        
        all_price_types: List[str] = instrument_manager.config["options"]["price_types"]
        price_columns = [px_type.capitalize() for px_type in all_price_types]
        if "Mid" in price_columns:
            init_other_columns = ["IVM"]
        
        total_columns = price_columns + [f"IV{px_type[0]}" for px_type in price_columns]
        

        self.main_view: MainView = MainView(instrument_manager=instrument_manager,
                                            tick_label_engine=tick_label_engine,
                                            metricLabels=price_columns + init_other_columns,
                                            **kwargs
                                            )
        
        self.options_container=instrument_manager.options_instrument_container
        underlying_objects = self._collect_underlying_objects(instrument_manager)
        spot_objects = self._spot_objects(instrument_manager.spot_instrument_container)
        if len(underlying_objects) == 1:
            self.widget_settings = settings.Settings(self.main_view.strike_table_column,
                                                     spot_objects=spot_objects,
                                                     columnCollection={"default" : price_columns + init_other_columns,
                                                                       "total" : total_columns},
                                                     )
            expiry_underlying_map=None
        else:
            expiry_underlying_map={}
            for underlying_obj in underlying_objects:
                if hasattr(underlying_obj, "expiry"):
                    expiry = underlying_obj.expiry
                    expiry_underlying_map[expiry] = underlying_obj
                    
            self.main_view.calls_table.add_expiry_underlying_map(expiry_underlying_map)
            self.main_view.puts_table.add_expiry_underlying_map(expiry_underlying_map)
            
            self.widget_settings = settings.Settings(self.main_view.strike_table_column,
                                                     spot_objects=spot_objects)

        self.main_layout_v = QtWidgets.QVBoxLayout(self)
        self.widget_settings.setSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                           QtWidgets.QSizePolicy.Fixed)

        self.setup_default(instrument_manager.spot_instrument_container.objects)
        
        self.main_layout_v.addWidget(self.widget_settings)
        self.main_layout_v.addWidget(self.main_view)
        self.setLayout(self.main_layout_v)
        
        self.main_view.calls_table.right_click_menu.setParent(self)
        self.main_view.puts_table.right_click_menu.setParent(self)
    
    def setup_default(self, spot_objects):
        default_n_strikes = 5
        
        all_strikes = list(self.options_container.maps.strike_expiry_map.keys())
        if len(spot_objects) > 0:
            self.spot_object = list(spot_objects.values())[0]
            if not np.isnan(self.spot_object.mid):
                strike_center = self.spot_object.mid
            else:
                strike_center = 0.5 * (np.amin(all_strikes) + np.amax(all_strikes))
        else:
            strike_center = 0.5 * (np.amin(all_strikes) + np.amax(all_strikes))

        self.widget_settings.strike_center_line.setText(str(strike_center))
        self.widget_settings.strike_center_line.editingFinished.emit()
        self.widget_settings.n_strikes_line.setText(str(default_n_strikes))
        self.widget_settings.n_strikes_line.editingFinished.emit()

    def _collect_underlying_objects(self, instrument_manager: InstrumentManager):
        underlying_objects = []
        for option_object in instrument_manager.options_instrument_container.objects.values():
            underlying_ticker = option_object.underlying_ticker
            underlying_object = instrument_manager.all_instrument_objects[underlying_ticker]
            if not underlying_object in underlying_objects:
                underlying_objects.append(underlying_object)
        return underlying_objects

    def create_underlying_label(self, instrument_manager: InstrumentManager):
        underlying_objects = []
        for option_object in instrument_manager.options_instrument_container.objects.values():
            underlying_object = option_object.underlying_object.ticker
            if not underlying_object in underlying_objects:
                underlying_objects.append(underlying_object)
        
        if len(underlying_objects) == 1:
            self.spot_name = list(instrument_manager.spot_instrument_container.objects.keys())[0]
            self.spot_qlabel = QtWidgets.QLabel(f"{self.spot_name}: {self.spot_object.mid}")
            self.spot_qlabel.setStyleSheet(stylesheets.get_settings_stylesheets("SpotQLabel"))
            self.spot_qlabel.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        else:
            if len(instrument_manager.spot_instrument_container.objects) > 0:
                self.spot_name = list(instrument_manager.spot_instrument_container.objects.keys())[0]
                self.spot_object = list(instrument_manager.spot_instrument_container.objects.values())[0]
                
                self.spot_qlabel = QtWidgets.QLabel(f"{self.spot_name}: {self.spot_object.mid}")
                self.spot_qlabel.setStyleSheet(stylesheets.get_settings_stylesheets("SpotQLabel"))
                self.spot_qlabel.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

    def _spot_objects(self, spot_instrument_container):
        if not spot_instrument_container is None:
            return list(spot_instrument_container.objects.values())
    



from pyqtgraph.opengl import GLViewWidget, GLMeshItem, MeshData
import numpy as np

# Create a view
view = GLViewWidget()

# Define a cube mesh (or any custom shape)
cube = MeshData.cube()

# Positions: Nx3 array of points
positions = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

# Create instanced mesh item
mesh_item = GLMeshItem(
    meshdata=cube,
    smooth=False,
    shader="shaded",
    glOptions="additive",
)
mesh_item.setData(positions=positions)  # Set positions
view.addItem(mesh_item)

from pyqtgraph.opengl import GLScatterPlotItem

scatter = GLScatterPlotItem()
scatter.setData(pos=positions, size=0.5, pxMode=False)

# Use a custom shader (disable spherical discard)
scatter.__glOpts["shader"].shaders[0].frag = """
#version 150
in vec4 vColor;
out vec4 fragColor;
void main() {
    fragColor = vColor;  // Draws squares (no discard)
}
"""
scatter.update()  # Force shader update


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable

from datetime import datetime
import numpy as np
from .. import utils
from functools import partial
from . import functions


def get_attribute_label_maps():
    metric_label_map = {"delta" : "Delta",
                        "ivol" : "Implied Volatility",
                        "TVAR" : "Total Variance",
                        "variance" : "Variance",
                        "expiry" : "Expiry",
                        "years" : "Years",
                        "strike" : "Strike",
                        "moneyness" : "Moneyness",
                        "log_moneyness" : "Log-Moneyness" ,
                        "forward_moneyness" : "Forward Moneyness",
                        "standardised_moneyness" : "Standardised-Moneyness" ,
                        "VAR" : "Variance",
                        "TVOL" : "Total Volatility"
                        }
    
    label_metric_map = {label : metric for metric, label in metric_label_map.items()}
    label_metric_map["Date"] = "expiry"
    return metric_label_map, label_metric_map


class TickEngine:
    def __init__(self, label=None, **kwargs):
        self.label=label
        self._tick_functions = {"Expiry": self.Expiry_function,
                                "Date" : self.Date_function,
                                "Variance" : self.Variance_function,
                                "Years": self._null,
                                "Delta": self.Delta_function,
                                "Strike": self._null,
                                "Moneyness": self.Moneyness_func,
                                "Log-Moneyness": self._null,
                                "Standardised-Moneyness" : self._null,
                                "Forward Moneyness" : self._null,
                                "Implied Volatility": self._null,
                                "Total Volatility": self._null,
                                "Total Variance" : self._null,
                                "Variance" : self._null,
                                }
        
        self.function=self._tick_functions[self.label] if not self.label is None else None
    
    def get_ticks(self, values, n):
        if self.label == "Delta":
            self.function()
    
    def __call__(self, values):
        return self.function(values)
        
    def valid_data(self, flag):
        if flag:
            self.function = self._tick_functions[self.label]
        else:
            self.function = self._null
    
    def get_label(self):
        return self.label
    
    def get_function(self):
        return self.function
    
    def change_function(self, label):
        self.label=label
        self.function=self._tick_functions[label]
        
    @staticmethod
    def _null(values):
        return [f"{value:,.2f}" for value in values]

    @staticmethod
    def Moneyness_func(values):
        return [f"{value:,.2f}" for value in values]

    @staticmethod
    def null_visualise(values):
        return [f"{val:,.2f}" for val in values]
    
    @staticmethod
    def rounder(tick_labels):
        rounded = np.round(tick_labels)
        can_be_integer = np.all(np.isclose(tick_labels, rounded))
        if can_be_integer:
            return rounded.astype(int).astype(str).tolist()
        else:
            round_val = int(abs(np.floor(np.log10(abs(np.diff(tick_labels)).min()))))
            return np.round(tick_labels, round_val).astype(str).tolist()
     
    @staticmethod
    def Delta_function(values):
        if max(values) > 3:
            multiplier = 100
        else:
            multiplier = 1
        if values[0] < 0.5 * multiplier < values[-1]: 
            values = np.linspace(values[0], values[-1], len(values)).tolist()
            new_values=[]
            for value in values:
                if value < 0.5 * multiplier:
                    new_values.append(f"{round(value, 1)}P")
                elif value == 0.5 * multiplier:
                    new_values.append("ATM")
                else:
                    new_values.append(f"{round(1 - value, 2)}C")
            return new_values
        else:
            new_values = [f"{round(value, 1)}P" if value < 0.5 * multiplier else f"{round(1 - value, 2)}C" for value in values]
            return new_values

    @staticmethod
    def Variance_function(values):
        return [f"{round(value**2, 2)}" for value in values]
    
    @staticmethod
    def Expiry_function(values):
        return [datetime.fromtimestamp(value).strftime("%d-%b-%y").upper() if value == value and 0<value<16725186000 else str(value) for value in values]
    
    @staticmethod
    def Date_function(values):
        return TickEngine.Expiry_function(values)
            
    @staticmethod
    def Years_function(values):
        return [f"{val:,.2f}" for val in values]


class TickEngineManager:
    def __init__(self, x_label=None, y_label=None, z_label=None):
        self.x_engine=TickEngine(x_label) if not x_label is None else None
        self.y_engine=TickEngine(z_label) if not y_label is None else None
        self.z_engine=TickEngine(z_label) if not z_label is None else None

        self._int_str_map = {i : s for i, s in enumerate("xyz")}
            
    def add_valid_callbacks(self, data_container):
        data_container.add_valid_data_callback(self.x_engine.valid_data)
        data_container.add_valid_data_callback(self.y_engine.valid_data)
        if not self.z_engine is None:
            data_container.add_valid_data_callback(self.z_engine.valid_data)
    
    def change_function(self, axis_label, axis_direction):
        getattr(self, f"{axis_direction}_engine", axis_label).change_function(axis_label)
    
    def get_engine(self, axis_direction):
        if isinstance(axis_direction, str):
            return getattr(self, f"{axis_direction}_engine")
        elif isinstance(axis_direction, int):
            ax_str = self._int_str_map.get(axis_direction, None)
            return getattr(self, f"{ax_str}_engine")

def get_metric_maps():    
    metric_maps = {"delta" : ["OTM", "delta", "call_flag", "delta_mag"],
                   "moneyness" : ["moneyness"],
                   "log_moneyness" : ["log_moneyness"],
                   "forward_moneyness" : ["forward_moneyness"],
                   "standardised_moneyness" : ["standardised_moneyness"],
                   "strike" : [],
                   "expiry" : [],
                   "years" : [],
                   "ivol" : [],
                   "IVOL_perc" :[],
                   "TVAR" : [],
                   }
    
    return metric_maps

class MetricAxisEngine:
    def __init__(self, x_base, y_base, x_label, y_label, z_label):
        self.x_base = x_base.copy()
        self.y_base = y_base.copy()
        self.metric_label_map, self.label_metric_map = get_attribute_label_maps()      
        self.x_metric=self.label_metric_map[x_label]
        self.y_metric=self.label_metric_map[y_label]
        self.z_metric=self.label_metric_map[z_label]

        self.metric_functions = {"expiry": self.null_metric,
                                 "date": self.null_metric,
                                 "years": self.years_metric_func,
                                 "delta": self.delta_metric_mask_sorter,
                                 "strike": self.null_metric,
                                 "moneyness": self.moneyness_mask_sorter,
                                 "log_moneyness": self.log_moneyness_mask_sorter,
                                 "standardised_moneyness": self.standardised_moneyness_sorter,
                                 "forward_moneyness" : self.forward_moneyness_sorter,
                                 "ivol": self.null_metric,
                                 "TVAR": self.TVAR_function,
                                 "TVOL" : self.TVOL_function,
                                 "VAR" : self.VAR_function,
                                 }
        self.base_idx = np.arange(self.y_base.size, dtype=int)       
        
        self.x_function = self.metric_functions[self.x_metric]      
        self.y_function = self.metric_functions[self.y_metric]      
        self.z_function = self.metric_functions[self.z_metric]      
        self.x_singular = self.singular_null
        self.y_singular = self.singular_null
        self.z_singular = self.singular_null

    def singular_null(self, x):
        return x
        
    def transform_axis(self, axis_direction, args):
        return getattr(self, f"{axis_direction}_singular")(args)
    
    def transform_values(self,
                         x: np.ndarray,
                         y: np.ndarray,
                         ivol: np.ndarray=None,
                         **kwargs
                         ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:        
        x_orig, y_orig = x, y
        z_orig = ivol.copy()
        x_metric, y, z, mask_removal_x, mask_rearrange_x = self.x_function(x=x_orig, y=y_orig, z=z_orig, **kwargs)
        
        x_new = x_orig[mask_removal_x][mask_rearrange_x]
        
        x_new, y_metric, z, mask_removal_y, mask_rearrange_y = self.y_function(x=x_new, y=y, z=z, **kwargs)
        
        x_metric = x_metric[mask_removal_y][mask_rearrange_y]
        y_new = y_orig[mask_removal_x][mask_rearrange_x][mask_removal_y][mask_rearrange_y]
        
        x, y, z_metric, mask_removal_z, mask_rearrange_z = self.z_function(x=x_new, y=y_new, z=z, **kwargs)
        
        x_metric = x_metric[mask_removal_z][mask_rearrange_z]
        y_metric = y_metric[mask_removal_z][mask_rearrange_z]
        
        idx_map = self.base_idx[mask_removal_x][mask_rearrange_x][mask_removal_y][mask_rearrange_y][mask_removal_z][mask_rearrange_z]
        
        return x_metric, y_metric, z_metric, idx_map

    def get_metric(self, axis_direction):
        return getattr(self, f"{axis_direction}_metric")
    
    def get_function(self, axis_direction):
        return getattr(self, f"{axis_direction}_function")
    
    def change_function(self, label, axis_direction):
        metric = self.label_metric_map[label]
        setattr(self, f"{axis_direction}_metric", metric)
        setattr(self, f"{axis_direction}_function", self.metric_functions[metric])
                   
    @staticmethod
    def null_metric(x=None, y=None, z=None, **kwargs):
        return x, y, z, [True] * z.size, np.arange(z.size)
    
    @staticmethod
    def _base_money_sorter(x=None, y=None, z=None, OTM=None, **kwargs):
        mask_removal = OTM        
        if not y is None:
            x, y, z = x[mask_removal], y[mask_removal], z[mask_removal]
            mask_rearrange = np.lexsort((x, y))
            return x[mask_rearrange], y[mask_rearrange], z[mask_rearrange], mask_removal, mask_rearrange
        else:
            x, z = np.where(mask_removal, x, np.nan), np.where(mask_removal, z, np.nan)
            ind = np.argsort(x, axis=0)
            return x, None, z, mask_removal, ind

    @staticmethod
    def TVAR_function(x=None, y=None, z=None, **kwargs):
        return x, y, (z**2) * utils.convert_unix_maturity_to_years(y), [True] * z.size, np.arange(z.size)

    @staticmethod
    def TVOL_function(x=None, y=None, z=None, **kwargs):
        return x, y, z * utils.convert_unix_maturity_to_years(y), [True] * z.size, np.arange(z.size)
    
    @staticmethod
    def VAR_function(x=None, y=None, z=None, **kwargs):
        return x, y, z**2, [True] * z.size, np.arange(z.size)

    @staticmethod
    def IVOL_perc_function(x=None, y=None, z=None, **kwargs):
        return x, y, 100 * z, [True] * z.size, np.arange(z.size)
    
    def moneyness_mask_sorter(self, x=None, y=None, z=None, moneyness=None, OTM=None, **kwargs):
        x = moneyness
        return self._base_money_sorter(x=x, y=y, z=z, OTM=OTM)
    
    def log_moneyness_mask_sorter(self, x=None, y=None, z=None, log_moneyness=None, OTM=None, **kwargs):
        x = log_moneyness
        return self._base_money_sorter(x=x, y=y, z=z, OTM=OTM)
    
    def standardised_moneyness_sorter(self, x=None, y=None, z=None, standardised_moneyness=None, OTM=None, **kwargs):
        x = standardised_moneyness
        return self._base_money_sorter(x=x, y=y, z=z, OTM=OTM)
    
    def forward_moneyness_sorter(self, x=None, y=None, z=None, forward_moneyness=None, OTM=None, **kwargs):
        x = forward_moneyness
        return self._base_money_sorter(x=x, y=y, z=z, OTM=OTM)

    @staticmethod
    def years_metric_func(x=None, y=None, z=None, **kwargs):
        return x, utils.convert_unix_maturity_to_years(y), z, [True] * z.size, np.arange(z.size)

    def delta_metric_mask_sorter(self, x=None, y=None, z=None, flatten=False, OTM=None, delta=None, call_flag=None, **kwargs):
        delta_mag=np.abs(delta)
        if flatten:
            mask_removal = OTM.flatten() & (delta_mag.flatten() < 0.5)
            put_indices = np.where(~call_flag.flatten()[mask_removal])[0]
            call_indices = np.where(call_flag.flatten()[mask_removal])[0]
            x_masked, y_masked, z_masked = delta.flatten()[mask_removal], y[mask_removal], z[mask_removal]
        else:
            mask_removal = OTM & (delta_mag < 0.5)
            put_indices = np.where(~call_flag[mask_removal])[0]
            call_indices = np.where(call_flag[mask_removal])[0]
            x_masked, y_masked, z_masked = delta[mask_removal], y[mask_removal], z[mask_removal]
        
        sorted_put_indices = put_indices[np.lexsort((y_masked[put_indices], -x_masked[put_indices]))] if put_indices.size > 0 else np.array([], dtype=int)
        sorted_call_indices = call_indices[np.lexsort((y_masked[call_indices], -x_masked[call_indices]))] if call_indices.size > 0 else np.array([], dtype=int)

        mask_rearrange = np.concatenate([sorted_put_indices, sorted_call_indices])
        x_sorted, y_sorted, z_sorted = x_masked[mask_rearrange], y_masked[mask_rearrange], z_masked[mask_rearrange]
        mask = x_sorted > 0
        x_sorted[~mask] = -x_sorted[~mask]
        x_sorted[mask] = 1 - x_sorted[mask]
        return x_sorted, y_sorted, z_sorted, mask_removal, mask_rearrange




from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any

from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget

if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GLViewBox import GLViewBox
import importlib
from pprint import pprint   
from OpenGL import GL
from OpenGL.GL import shaders
import numpy as np

from pyqtgraph.Qt import QtGui, QT_LIB
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

__all__ = ['GLLinePlotItem']
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from pyqtgraph.Qt import QT_LIB
import importlib
import pyqtgraph as pg
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DViewBox
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DViewBox
from pyqtgraph import opengl
import sys
import cProfile
import sys
import pstats
from pyqtgraph import ButtonItem, icons
QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

from PySide6.QtTest import QTest
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph.opengl import GLLinePlotItem, GLScatterPlotItem
from pyqtgraph import Transform3D
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DLinePlotDataItem



class CustomGLLinePlotItem(GLLinePlotItem):
    def __init__(self, *args, **kwargs):
        self._updatesBlocked=False
        super().__init__(*args, **kwargs)   
        
    def updatesBlocked(self):
        return self._updatesBlocked
    
    def update(self):
        if not self.updatesBlocked():
            super().update()
    
    def blockUpdates(self, flag):
        self._updatesBlocked=flag


class TickLineset(CustomGLLinePlotItem):
    def __init__(self, *args, **kwargs):
        mode = "lines"    # Keep mode lines for optimized painting for grouped line items 
        kwargs["mode"]=mode
        self._canPaint=True
        self._lineItems: List[GLLinePlotItem]=[]
        self._linkedLineItems: List[GLLinePlotItem]=[]
        self._zOffset=0
        
        self._contantArray: np.ndarray=None
        self._linkeLineset: TickLineset=None
        self._lineItemsContantMap: Dict[GLLinePlotItem, int] = {}
        self._linkedLinesetsConstantMap: Dict[TickLineset, Dict[GLLinePlotItem, Tuple[int, int]]] = {}
        
        super().__init__(*args, **kwargs)

    def stackLineset(self, newPositionAll, lineItems: List[GLLinePlotItem]):
        for item in lineItems:
            if item.transform().isIdentity():
                newPos = item.pos
            else:
                tr = np.array(item.transform().data()).reshape(4, 4).T  
                points_h = np.hstack([item.pos, np.ones((item.pos.shape[0], 1))])
                transformed_h = points_h @ tr.T
                newPos = transformed_h[:, :3]
            
            if newPositionAll is None:
                newPositionAll=newPos
            else:
                newPositionAll = np.vstack((newPositionAll, newPos))
        return newPositionAll
    
    def lineItems(self):
        return self._lineItems
    
    def addLineItem(self, item: GLLinePlotItem):        
        self._lineItems.append(item)
        
    def linkLineset(self, lineset):
        self._linkedLineItems = lineset
        
        
    def paint(self):
        print("paint")
        if not self._canPaint:
            return

        combined_pos = None
        
        # Process direct line items with z-offset
        for item in self._lineItems:
            pos = self._getWorldPositions(item)
            if pos is not None:
                pos[:, 2] += self._zOffset  # Apply z-offset
                combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
        
        for lineset in self._linkedLineItems:
            for item in lineset.lineItems():
                pos = self._getWorldPositions(item)
            
                #pos = item.pos
                if pos is not None:
                    pos[:, 2] -= self._zOffset
                    combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
        print(np.shape(combined_pos))

        if combined_pos is not None:
            self.blockUpdates(True)
            self.setData(pos=combined_pos)
            self.blockUpdates(False)
            
        super().paint()

    def _getWorldPositions(self, item):
      #  self.itemTr
        """Convert item's positions to this lineset's coordinate system"""
        if not hasattr(item, 'pos') or item.pos is None:
            return None
        
        tr = item.transform()
        pos = item.pos.copy()  # Create a copy to avoid modifying original
        
        if not tr.isIdentity():
            points_h = np.hstack([pos, np.ones((pos.shape[0], 1))])
            tr_matrix = np.array(tr.data()).reshape(4, 4).T
            transformed_h = points_h @ tr_matrix.T
            pos = transformed_h[:, :3]
        
        return pos



class Custom(QtWidgets.QGraphicsObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)   

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
        self.gl_widget = opengl.GLViewWidget()
        
        self.setCentralWidget(self.gl_widget)
        
        
        self.group1 = GLGraphicsItem()
        self.group1.translate(0, 1, 1)
        self.group2 = GLGraphicsItem()
        self.group2.translate(1, 0, -1)

        
        self.lineset1 = TickLineset(mode="lines", color="cyan")
        self.lineset1._zOffset=0.1
        self.lineset2 = TickLineset(mode="lines", color="blue")
        
        
     
        
        
        self.group4 = TickLineset()
        lines = []

        num_lines = 5
        start, end = 0, 1
        length = end - start
        positions = np.linspace(start, end + 2, num_lines)
        

    #    self.lineset1.translate(-0.5, 0.5, 0)
    #    self.lineset2.translate(-0.5, -.5, 0)


        for x in [start, end][:1]:
            for z in positions:
                pos = np.array([[x, start, z], [x, end, z]])
                line1 = opengl.GLLinePlotItem(pos=pos)

                
                line1.translate(0, 1, 0)
                
                self.lineset1.addLineItem(line1)
                
                
                
                line1.setParentItem(self.lineset1)
                
                
                line2 = opengl.GLLinePlotItem(pos=pos, color="blue")


                
                line2.translate(0.6, 0, 0)

                
                self.lineset2.addLineItem(line2)

                
                
                
                line2.setParentItem(self.lineset2)

        self.lineset1.setParentItem(self.group1)
        self.lineset2.setParentItem(self.group2)


        self.lineset1.linkLineset([self.lineset2])
        
        
        def _null():return
        
        self.group2.paint=_null
        
        pos = np.array(pos)
        
        

        self.gl_widget.addItem(self.group1)

        
        for i in range(1, 11):
            txt = opengl.GLTextItem(pos = (i/10, 0, 0), text=str(i))
            self.gl_widget.addItem(txt)
            txt = opengl.GLTextItem(pos = (0, i/10, 0), text=str(i))
            self.gl_widget.addItem(txt)
            txt = opengl.GLTextItem(pos = ( 0, 0, i/10), text=str(i))
            self.gl_widget.addItem(txt)
        
        
        
        self.showMaximized()
        self.update()
                
        
        
    def update_timer(self):


        #self.group.rotate(90., 0., 0., 1.)
        grid_pos = self.gridline.pos
        grid_pos[:, 1] = grid_pos[:, 1] + 0.2
        ticks_pos = self.ticks.pos
        ticks_pos[:, 1] = ticks_pos[:, 1]  + 0.2

        self.gridline.setData(pos=grid_pos)
        self.ticks.setData(pos=ticks_pos)
        
        
        
        
        self.lineset2.update()
        
        


def main():
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    #timer = QtCore.QTimer()
    #timer.timeout.connect(win.update_timer)
    #timer.start(300)


    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from PySide6.QtWidgets import QApplication, QMainWindow, QTableWidget, QTableWidgetItem

app = QApplication([])

# Create the main window
main_window = QMainWindow()
main_window.setWindowTitle("Table as Central Widget Example")

# Create a QTableWidget
table = QTableWidget()
table.setRowCount(3)  # 3 rows
table.setColumnCount(2)  # 2 columns

# Fill the table with some data
table.setHorizontalHeaderLabels(["Name", "Age"])

table.setItem(0, 0, QTableWidgetItem("Alice"))
table.setItem(0, 1, QTableWidgetItem("25"))
table.setItem(1, 0, QTableWidgetItem("Bob"))
table.setItem(1, 1, QTableWidgetItem("30"))
table.setItem(2, 0, QTableWidgetItem("Charlie"))
table.setItem(2, 1, QTableWidgetItem("35"))

# Set the table as the central widget
main_window.setCentralWidget(table)

main_window.show()
app.exec()


s# type: ignore
# ruff: noqa: F821

import time
import os
from pprint import pprint

import subdir.profiling_test_script2 as script2


@profile
def fact(n):
    result = 1
    for i in range(2, n // 4):
        result *= i
    result = 1
    # This is a comment
    for i in range(2, n // 16):
        result *= i
    result = 1

    if False:
        # This won't be run
        raise RuntimeError("What are you doing here ???")

    for i in range(2, n + 1):
        result *= i
    return result
    # This is after the end of the function.

    if False:
        # This won't be run
        raise RuntimeError("It's getting bad.")


@profile
def wont_run(n):
    # Don't try this at home!
    while True:
        time.sleep(1)


@profile
def sum_(
    n,
    this_very_long_argument_list,
    and_this_one,
    and_another_one,
    this_is_the_last_one_i_promise,
    ___i_m_sorry,
):
    """Docstring"""
    # Comment
    result = 0
    "string"

    for i in range(1, n + 1):
        result += i
    return result


if __name__ == "__main__":
    print(fact(120))
    print(sum_(120, 0, 0, 0, 0, 0))
    print(script2.fact2(120))
    print(script2.sum2(120))
    pprint(dict(os.environ))


"""
Demonstrates use of GLGraphItem
"""

import sys

import numpy as np

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui
import pyqtgraph.opengl as gl

if 'darwin' in sys.platform:
    fmt = QtGui.QSurfaceFormat()
    fmt.setRenderableType(fmt.RenderableType.OpenGL)
    fmt.setProfile(fmt.OpenGLContextProfile.CoreProfile)
    fmt.setVersion(4, 1)
    QtGui.QSurfaceFormat.setDefaultFormat(fmt)

app = pg.mkQApp("GLGraphItem Example")
w = gl.GLViewWidget()
w.setCameraPosition(distance=5)
w.show()

edges = np.array([
    [0, 2],
    [0, 3],
    [1, 2],
    [1, 3],
    [2, 3]
])

nodes = np.array(
    [
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 1]
    ]
)

edgeColor=pg.glColor("w")

gi = gl.GLGraphItem(
    edges=edges,
    nodePositions=nodes,
    edgeWidth=1.,
    nodeSize=0.1,
    pxMode=False
)

w.addItem(gi)

if __name__ == "__main__":
    pg.exec()



#%%
from pyqtgraph.opengl import GLGraphicsItem
from PySide6 import QtCore, QtGui, QtWidgets
from pyqtgraph.opengl import GLViewWidget

import abc
from abc import ABC, ABCMeta, abstractmethod
from PySide6.QtCore import QObject

    
import abc
from PySide6.QtCore import QObject

class QABCMeta(abc.ABCMeta, type(QObject)):
    
    """
    SOURCE:
    https://stackoverflow.com/questions/78778926/creating-a-metaclass-that-inherits-from-abcmeta-and-qobject
    
    """
    
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)






class _ABCQWidgetMeta(type(QtCore.QObject), ABCMeta):...
class ABCQWidget(QtWidgets.QTableWidget, metaclass=_ABCQWidgetMeta):...


class CustomTableWidget(QtWidgets.QWidget, metaclass=ABCQWidget):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)


app = QtWidgets.QApplication()
a = CustomTableWidget()



import sys
import numpy as np

from vispy import app, scene
from vispy.util.filter import gaussian_filter


canvas = scene.SceneCanvas(keys='interactive', bgcolor='w')
view = canvas.central_widget.add_view()
view.camera = scene.TurntableCamera(up='z', fov=60)


z = np.random.normal(size=(30, 30), scale=200)
z[15, 15] += 50000
z = gaussian_filter(z, (10, 10))
p1 = scene.visuals.SurfacePlot(z=z, color=(0.3, 0.3, 1, 1))
p1.transform = scene.transforms.MatrixTransform()
p1.transform.scale([1/249., 1/249., 1/249.])
p1.transform.translate([-0.5, -0.5, 0])

view.add(p1)

# p1._update_data()  # cheating.
# cf = scene.filters.ZColormapFilter('fire', zrange=(z.max(), z.min()))
# p1.attach(cf)


xax = scene.Axis(pos=[[-0.5, -0.5], [0.5, -0.5]], tick_direction=(0, -1),
                 font_size=16, axis_color='k', tick_color='k', text_color='k',
                 parent=view.scene)
xax.transform = scene.STTransform(translate=(0, 0, -0.2))

yax = scene.Axis(pos=[[-0.5, -0.5], [-0.5, 0.5]], tick_direction=(-1, 0),
                 font_size=16, axis_color='k', tick_color='k', text_color='k',
                 parent=view.scene)
yax.transform = scene.STTransform(translate=(0, 0, -0.2))

# Add a 3D axis to keep us oriented
axis = scene.visuals.XYZAxis(parent=view.scene)

if __name__ == '__main__':
    canvas.show()
    if sys.flags.interactive == 0:
        app.run()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, ClassVar
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox 

import weakref
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing.meta import QABCMeta, abc
from pyqtgraph.opengl import GLGraphicsItem
from .AbstractGLGraphicsItem import AbstractGLGraphicsItem
from .AbstractGLGraphicsItem import ABCGraphicsItemMeta
from PySide6 import QtCore, QtWidgets
import numpy as np
from dataclasses import dataclass, field, InitVar
from abc import ABC, abstractmethod
import warnings
from pyqtgraph.Qt import QT_LIB
from functools import wraps 
import traceback
from pprint import pprint
from pyqtgraph import opengl



@dataclass(slots=True)
class AbstractPlotDataset(ABC):
    initData: InitVar[np.ndarray|Tuple[np.ndarray, ...]] = None
    allFinite: List[Optional[bool]] = field(default_factory=list)
    _dataBounds: List[List[float]] | None = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._initialize_attributes()
        
        
    def _initialize_attributes(self):
        if len(self.allFinite) == 0:
            self.allFinite = [None] * self.num_components
            self._updateDataBounds()
        
    def dataBounds(self) -> QtCore.QRectF | None:
        if self._dataBounds is None: 
            self._updateDataBounds()
        return self._dataBounds
    
    @abstractmethod
    def data(self) -> np.ndarray | Tuple[np.ndarray,...]:
        ...
        
    @abstractmethod
    def dataKwargs(self) -> Dict[str, np.ndarray]:
        ...

    @property
    @abstractmethod
    def num_components(self) -> int:
        """Number of data components (e.g., 3 for x/y/z)"""
        
    @abstractmethod
    def getComponents(self) -> List[np.ndarray]:
        """Get list of component arrays (flattened for bounds calculation)"""
    
    def _updateDataBounds(self):
        components = self.getComponents()
        if any([values is None for values in components]):
            return None
         
        dataBounds=[]
        for idx, arr in enumerate(components):
            minRange, maxRange, finiteFlag = self._getArrayBounds(arr, self.allFinite[idx])
            dataBounds.append((minRange, maxRange))
            self.allFinite[idx]=finiteFlag
        self._dataBounds=dataBounds
                
    def _getArrayBounds(self,
                        arr: np.ndarray,
                        all_finite: bool | None
                        ) -> tuple[float, float, bool]:
        # here all_finite could be [None, False, True]
        if not all_finite:  # This may contain NaN or inf values.
            # We are looking for the bounds of the plottable data set. Infinite and Nan
            # are ignored.
            selection = np.isfinite(arr)
            # True if all values are finite, False if there are any non-finites
            all_finite = bool(selection.all())
            if not all_finite:
                arr = arr[selection]
        
        # here all_finite could be [False, True]
        try:
            amin = np.min( arr )  # find minimum of all finite values
            amax = np.max( arr )  # find maximum of all finite values
        except ValueError:  # is raised when there are no finite values
            amin = np.nan
            amax = np.nan
        return amin, amax, all_finite

@dataclass(slots=True)
class PlotDatasetFlatMesh(AbstractPlotDataset):    
    _pos: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._pos = initData
        super(PlotDatasetFlatMesh, self).__post_init__(initData)
        
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        return [self._pos[:, i] for i in range(self.num_components)]
    
    @property
    def pos(self) -> np.ndarray:
        return self._pos
        
    @pos.setter
    def pos(self, value: np.ndarray):
        self._pos = value

    def data(self) -> np.ndarray:
        return self.pos
    
    def dataKwargs(self):
        return {"pos" : self.pos}
            
@dataclass(slots=True)
class PlotDatasetMesh(AbstractPlotDataset):
    initData: InitVar[Tuple[np.ndarray, ...]] = None
    
    _x: np.ndarray = field(init=False, default=None)
    _y: np.ndarray = field(init=False, default=None)
    _z: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self.x, self.y, self.z = initData
        super(PlotDatasetMesh, self).__post_init__(initData)
            
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        return [c.ravel() for c in (self.x, self.y, self.z)]
    
    # Property setters with automatic bounds update
    @property
    def x(self) -> np.ndarray:
        return self._x
        
    @x.setter
    def x(self, value: np.ndarray):
        self._x = value
    
    @property
    def y(self) -> np.ndarray:
        return self._y
        
    @y.setter
    def y(self, value: np.ndarray):
        self._y = value
    
    @property
    def z(self) -> np.ndarray:
        return self._z
        
    @z.setter
    def z(self, value: np.ndarray):
        self._z = value

    def data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return (self.x, self.y, self.z)

    def dataKwargs(self):
        return {"x" : self.x, "y" : self.y, "z" : self.z}



@dataclass(slots=True)
class CacheData:
    _null3: ClassVar[List[None]] = [None, None, None]
    _false3: ClassVar[List[None]] = [False, False, False]
    _null32: ClassVar[List[List[None]]] = [[None, None],[None, None], [None, None]]
    
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _bounds: List[List[None]]|List[List[float]] = field(init=False, default_factory=lambda : [row.copy() for row in CacheData._null32])
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _orthoRange: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _frac: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _cacheFlag: List[bool] = field(init=False, default_factory=lambda : CacheData._false3.copy())

    def __repr__(self):
        out = {}
        for name in self.__slots__:
            if name.startswith('_'):
                public_name = name[1:]
                method = getattr(self, public_name, None)
                if callable(method):
                    try:
                        out[public_name] = method()
                    except Exception:
                        out[public_name] = "<error>"
        return f"{self.__class__.__name__}({out})"   
     
    def cacheFlag(self, ax=None):
        if ax is None:
            return self._cacheFlag
        else:
            return self._cacheFlag[ax]
        
    
    def setCacheFlag(self, flag, ax=None):
        if ax is None:
            self._cacheFlag=flag
        else:
            self._cacheFlag[flag] = ax

    def setBounds(self, bounds, ax=None):
        if ax is None:
            self._bounds = bounds
            self.setCacheFlag([not None in lim for lim in bounds])
        else:
            self._bounds[ax] = bounds
            self.setCacheFlag(bounds, ax)

    
    def bounds(self, ax=None):
        if ax is None:
            return self._bounds
        else:
            return self._bounds[ax]

    
    def setMask(self, mask, ax=None):
        if ax is None:
            self._mask=mask     
        else:
            self._mask[ax] = mask
    
        
    def mask(self, ax=None):
        if ax is None:
            self._mask
        else:
            return self._mask[ax]


    def setOrthoRange(self, orthoRange, ax=None):
        if ax is None:
            self._orthoRange=orthoRange
        else:
            self._orthoRange[ax] = orthoRange

    def orthoRange(self, ax=None):
        if ax is None:
            return self._orthoRange
        else:
            return self._orthoRange[ax] 


    def setFrac(self, frac, ax=None):
        if ax is None:
            self._frac=frac
        else:
            self._frac[ax] = frac
        
    def frac(self, ax=None):
        if ax is None:
            return self._frac
        else:
            return self._frac[ax] 

    def reset(self):
        self.setBounds([row.copy() for row in self._null32])
        self.setCacheFlag(self._false3.copy())
        self.setMask(self._null3.copy())
        self.setOrthoRange(self._null3.copy())
        self.setFrac(self._null3.copy())



class MeshPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._x: np.ndarray
        self._y: np.ndarray 
        self._z: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def x(self):
        return self._x
      
    @x.setter
    def _x(self, value):
        self._x = value
        
    @property
    def y(self):
        return self._y
      
    @y.setter
    def _y(self, value):
        self._y = value

    @property
    def z(self):
        return self._z
      
    @z.setter
    def _z(self, value):
        self._z = value

class VectorPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._pos: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def pos(self):
        
        return self._pos
      
    @pos.setter
    def _pos(self, value):
        if (self.pos != value).any():
            
            self._pos = value


class BaseAbstractPlotDataItem(AbstractGLGraphicsItem, metaclass=ABCGraphicsItemMeta):
    sigPlotChanged = QtCore.Signal(object) 
    def __init__(self, topology_type="", dataKwargs=[], view_box=None, parentItem=None, **kwargs):
        self._dataset: AbstractPlotDataset=None
        self._datasetDisplay: AbstractPlotDataset=None
        self._cacheData = CacheData()
        self.plotdataset_cls=None
        self._dataKwargs=dataKwargs
        self._viewBox: GL3DViewBox=view_box
        if topology_type=="Mesh":
            self.plotdataset_cls=PlotDatasetMesh
        elif topology_type=="FlatMesh":
            self.plotdataset_cls=PlotDatasetFlatMesh
        else:
            traceback.print_stack()
            raise
        self.extra_opts={"clipToView" : False,
                         }

        self.setProperty('xViewRangeWasChanged', False)
        self.setProperty('yViewRangeWasChanged', False)
        self.setProperty('zViewRangeWasChanged', False)
        self.setProperty('styleWasChanged', True)  # force initial update
        
        self.setParentItem(parentItem)

    def viewTransform(self):
        view = self.getViewBox()
        if view is None:
            return None
        if hasattr(view, 'implements') and view.implements('ViewBox'):
            return self.itemTransform(view.innerSceneItem())[0]
        else:
            return self.sceneTransform()

    def setViewBox(self, vb):
        self._viewBox=vb

    def dataRect(self): 
        return None if self._dataset is None else self._dataset.dataBounds()

    def setClipToView(self, state: bool):
        pass
    
    def viewTransformChanged(self):
        self.invalidateBounds()
        #self.prepareGeometryChange()

    def invalidateBounds(self):
        self._cacheData.reset()
        
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):

        # view range has changed; re-plot if needed 
        update_needed = False
        if changed is None or changed[0]: 
            self.setProperty('xViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[1]:
            self.setProperty('yViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[2]:
            self.setProperty('zViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True

        if update_needed:
            self.updateItems(styleUpdate=False)
            
    def updateItems(self, styleUpdate: bool = True):
        dataset = self._getDisplayDataset()
        
        if dataset is None:
            print(f"\n\nDataSet is None!!!\n\n")
            raise
            self.hide()
            return
        else:
            self.updateData()
    
    def updateData(self, **kwargs):
        self.invalidateBounds()
        #self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.sigPlotChanged.emit(self)
    
    def getData(self):
        dataset = self._getDisplayDataset()
        return (None, None, None) if dataset is None else dataset.data()
            
    def getOriginalDataset(self) -> tuple[None, None] | tuple[np.ndarray, np.ndarray]:
        return self._dataset.data()
    
    def _getDisplayDataset(self) -> AbstractPlotDataset | None:
        if self._dataset is None:
            return None
        # Return cached processed dataset if available and still valid:
        if all((not self._datasetDisplay is None,
                not self.property('xViewRangeWasChanged'),
                not self.property('yViewRangeWasChanged'),
                not self.property('zViewRangeWasChanged')
                )):
            return self._datasetDisplay

        allFinite = self._dataset.allFinite
        data = self._dataset.data()
        
        view = self.getViewBox()
        
        if view is None:
            view_range = None
        else:
            view_range = view.viewRange()  # this is always up-to-date
        if view_range is None:
            view_range = self.viewRect()
 
        if view is None or view.autoRangeEnabled()[0]:
            pass  # no ViewBox to clip to, or view will autoscale to data range.
                
        if not all(allFinite):
            data = self._dataset.data()
        if not view_range is None and all(allFinite):
            data = self.clipDataFromVRange(view_range, data)        

        self._datasetDisplay = self.plotdataset_cls(data, allFinite)
        
        self.property('xViewRangeWasChanged', False)
        self.property('yViewRangeWasChanged', False)
        self.property('zViewRangeWasChanged', False)
        return self._datasetDisplay


    def dataBoundsVectorized(self,
                             frac=(1., 1. ,1),
                             orthoRange=(None, None, None)
                             ) -> Tuple[float, float, float]:

        if not self.visible() or self._dataset is None:
            return [[None, None], [None, None], [None, None]]
        else:
            data = self._dataset.data()
            mask = np.ones(len(data), dtype=bool)

            cache_bounds = {}
            mask_container={}

            for ax in range(2):
                if self._cacheData.cacheFlag(ax) is None and all((self._cacheData.frac(ax) == frac[ax],
                                                                  self._cacheData.orthoRange(ax) == orthoRange[ax])):
                    cache_bounds[ax] = self._cacheData.bounds(ax)
                    mask_container[ax] = self._cacheData.mask(ax)

                if not orthoRange[ax] is None:
                    mask = (data[:, ax] >= orthoRange[ax][0]) & (data[:, ax] <= orthoRange[ax][1])
                    self._cacheData.setMask(mask, ax)
                    mask_container[ax] = mask
                else:
                    self._cacheData.setMask(None, ax)
            
            x_filtered = data[0][mask_container[0]] if 0 in mask_container else data[0]
            y_filtered = data[1][mask_container[1]] if 1 in mask_container else data[1]
        
            if self._cacheData.cacheFlag(2) is None and all((self._cacheData.frac(2) == frac[2],
                                                             self._cacheData.orthoRange(2) == orthoRange[2]
                                                             )):
                cache_bounds[2] = self._cacheData.bounds(2)
                
                
            z_filtered = data[2]
            if 0 in mask_container:
                z_filtered = z_filtered[mask_container[0], :]
            
            if 1 in mask_container:
                z_filtered = z_filtered[:, mask_container[1]]

            if not orthoRange[2] is None:
                z_filtered = np.clip(z_filtered, orthoRange[2][0], orthoRange[2][1])

            
            limits = [[bound.min(), bound.max()] if i not in cache_bounds else cache_bounds[i] for i, bound in enumerate((x_filtered, y_filtered, z_filtered))]
            self._cacheData.setBounds(limits)
            return self._cacheData.bounds()

    @classmethod
    @abstractmethod
    def clipDataFromVRange(cls, dataset, viewrange):...
    
    @abstractmethod
    def _setDataChild(self, *args, **kwargs):...
        
        
    def setData(self, *args, **kwargs):
        self.blockUpdates(True)
        self._setDataChild(*args, **kwargs)
        self.updateItems(styleUpdate=self.property('styleWasChanged'))
        self.blockUpdates(False)
        self.update()

    def _paintHelper(self):
        data = self._compareDataUpdate()
        dataKwargs = {key : value for key, value in zip(self._dataKwargs, data)}
        
        self.blockUpdates(True)
        self._setDataChild(**dataKwargs)
        self.blockUpdates(False)

    @abstractmethod
    def _compareDataUpdate(self):...


class _Temp(BaseAbstractPlotDataItem):
    pass

class _ABCPlotDataItemMeta(ABCGraphicsItemMeta):
    def __call__(cls, *args, **kwargs):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        self = cls.__new__(cls, *args, **kwargs)
        self._original_class = cls
        self.__class__ = _Temp
        
        topology_type = cls.topology_type
        dataKwargs = cls.dataKwargs
        view_box = kwargs.pop("view_box", None)
        parentItem = kwargs.pop("parentItem", None)

        BaseAbstractPlotDataItem.__init__(self, topology_type=topology_type, dataKwargs=dataKwargs, parentItem=parentItem, view_box=view_box, )
        
        self.__class__ = self._original_class
        del self._original_class
        
        if hasattr(self, '__init__') and self.__init__ is not None:
            self.__init__(*args, **kwargs)
            
        return self

class BaseMeshPlotDataItem(BaseAbstractPlotDataItem, metaclass=_ABCPlotDataItemMeta):
    topology_type="Mesh"
    dataKwargs=["x", "y", "z"]

    def __init__(self, view_box=None, parentItem=None, **kwargs):
        self._x=None
        self._y=None
        self._z=None
        BaseAbstractPlotDataItem.__init__(self, view_box=view_box, parentItem=parentItem)
    
    def _compareDataUpdate(self):
        x, y, z = self.getData()
        if np.array_equal(self._x, x):
           x=None 
        if np.array_equal(self._y, y):
           y=None 
        if np.array_equal(self._z, z):
           z=None 
        return x, y, z

    @classmethod
    def clipDataFromVRange(cls, dataset, view_range):
        x, y, z = dataset
        
        x_min, x_max = view_range[0]
        y_min, y_max = view_range[1]
        z_min, z_max = view_range[2]
        
        x_mask = (x >= x_min) & (x <= x_max)
        y_mask = (y >= y_min) & (y <= y_max)
        
        x_filtered = x[x_mask]
        y_filtered = y[y_mask]
                
        z_filtered = z[x_mask, :][:, y_mask]
        z_filtered = np.clip(z_filtered, z_min, z_max)

        return x_filtered, y_filtered, z_filtered

class BaseMeshFlatPlotDataItem(BaseAbstractPlotDataItem, metaclass=_ABCPlotDataItemMeta):
    topology_type="FlatMesh"
    dataKwargs=["pos"]

    def __init__(self, view_box=None, parentItem=None, **kwargs):
        self.pos=None
        BaseAbstractPlotDataItem.__init__(self, view_box=view_box, parentItem=parentItem)
    
    
    def _compareDataUpdate(self):
        data = self.getData()
        if np.equal(self.pos, data):
            data=None

        return data
    
    @classmethod
    def clipDataFromVRange(cls, dataset, view_range):
        x, y, z = dataset
        
        x_min, x_max = view_range[0]
        y_min, y_max = view_range[1]
        z_min, z_max = view_range[2]
        
        x_mask = (x >= x_min) & (x <= x_max)
        y_mask = (y >= y_min) & (y <= y_max)
        
        x_filtered = x[x_mask]
        y_filtered = y[y_mask]
                
        z_filtered = z[x_mask, :][:, y_mask]
        z_filtered = np.clip(z_filtered, z_min, z_max)

        return x_filtered, y_filtered, z_filtered


class CustomGLSurfacePlotDataItem(BaseMeshPlotDataItem, opengl.GLSurfacePlotItem):
    def __init__(self, x=None, y=None, z=None, colors=None, parentItem=None, **kwds):
        BaseMeshPlotDataItem.__init__(self, parentItem=parentItem, x=x, y=y, z=z, **kwds)
    
    
    def _setDataChild(self, *args, **kwargs):
        return opengl.GLSurfacePlotItem.setData(self, *args, **kwargs)
    
    def _childPaint(self):
        return opengl.GLSurfacePlotItem.paint(self)

    
class CustomGLScatterPlotDataItem(BaseMeshFlatPlotDataItem, opengl.GLScatterPlotItem):
    def __init__(self, parentItem=None, **kwds):
        BaseMeshFlatPlotDataItem.__init__(self, parentItem=parentItem, **kwds)
    
    def _setDataChild(self, *args, **kwargs):
        return opengl.GLScatterPlotItem.setData(self, *args, **kwargs)
    
    def _childPaint(self):
        return opengl.GLScatterPlotItem.paint(self)


class CustomGraphPlotDataItem(BaseMeshFlatPlotDataItem, opengl.GLGraphItem):
    def __init__(self, parentItem=None, **kwds):
        BaseMeshFlatPlotDataItem.__init__(self, parentItem=parentItem, **kwds)
    
    def _setDataChild(self, *args, **kwargs):
        return opengl.GLGraphItem.setData(self, *args, **kwargs)
    
    def _childPaint(self):
        return opengl.GLGraphItem.paint(self)

class AbstractGLPlotItem:
    pass



from py_vollib.black.implied_volatility import implied_volatility
import py_vollib_vectorized

import time
price = 10.
S = 100.
K=102.
t=0.002
r=0.05
q=0.02
flag="c"




s =time.time()
for _ in range(20000):
    py_vollib_vectorized.vectorized_implied_volatility_black(price, S, K, r, t, flag)

print(time.time() - s)
print(py_vollib_vectorized.vectorized_implied_volatility_black(price, S, K, r, t, flag))





s =time.time()
for _ in range(20000):
    implied_volatility(price, S, K, r, t, flag)

print(time.time() - s)
print(implied_volatility(price, S, K, r, t, flag))



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox
    from ..widgets.GL3DViewWidget import GL3DViewWidget

    
from PySide6 import QtWidgets, QtCore, QtGui
from abc import ABC, abstractmethod, ABCMeta
import abc
from pyqtgraph import Transform3D, functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from PySide6.QtWidgets import QGraphicsItem
from enum import Enum, EnumType, Flag
import warnings
from pyqtgraph.Qt import QT_LIB
import weakref
from pyqtgraph.Transform3D import Transform3D


class QABCMeta(ABCMeta, type(QtCore.QObject)):
    """Meta class for OpenGL mixin objects, combining ABC and QObject."""
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)



qt_enums = [QGraphicsItem.GraphicsItemChange.ItemPositionChange,
            QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleChange, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemSceneChange, 
            QGraphicsItem.GraphicsItemChange.ItemTransformChange,
            QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemParentChange,
            QGraphicsItem.GraphicsItemChange.ItemParentHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemChildAddedChange,
            QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
            ]
    

class GLGraphicsEnumMetaclass(EnumType):
    def __new__(metacls, clsname, bases, clsdict):
        base_enums = clsdict.get('_base_enums', [])
        custom_members = clsdict.get('_custom_members', [])

        ignore_list = list(clsdict.get('_ignore_', []))
        ignore_list.extend(['_base_enums', '_custom_members'])
        clsdict['_ignore_'] = ignore_list

        values = [member.value for member in base_enums]
        max_value = max(values) if values else 0

        for member in base_enums:
            clsdict[member.name] = member.value

        shifts = [1, 2]
        for idx, (name, _) in enumerate(custom_members):
            shift = shifts[idx] if idx < len(shifts) else (idx + 1)
            value = max_value << shift
            clsdict[name] = value

        return super().__new__(metacls, clsname, bases, clsdict)
    
    def __call__(cls, value):
        return super().__call__(value)


qt_flags = [QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations,
            QGraphicsItem.GraphicsItemFlag.ItemHasNoContents,
            QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges,
            ]


class MetaFlag(type(Flag)):
    def __call__(cls, value=0):
        return super().__call__(value)


class GL3DGraphicsItemMixin(GLGraphicsItem, metaclass=QABCMeta):
    
    class GLGraphicsItemChange(Enum, metaclass=GLGraphicsEnumMetaclass):
        _ignore_ = ['_base_enums', '_custom_members'] 
        _base_enums = qt_enums
        _custom_members = [('ItemViewChange', None), ('ItemViewHasChanged', None),
                           ('ItemParentWidgetChange', None), ("ItemParentWidgetHasChanged", None),
                           ('ItemDataChange', None), ("ItemDataHasChanged", None),
                           ]

    class GLGraphicsItemFlag(Flag, metaclass=MetaFlag):
        NoFlag = 0
        ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
        ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
        ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value
        ItemSendsDataChanges = 0x100
        

        def __repr__(self):
            if self.value == 0:
                return f"{self.__class__.__name__}({self.value})"
            return super().__repr__()

    def __init__(self, parentItem:'GL3DGraphicsItemMixin'=None, **kwargs):
        self._flags = self.GLGraphicsItemFlag()
        self._blockUpdates: bool=False
        self._cachedView=None
        self._parentWidget=None
        self._connectedView=None
        self._explicitlyHidden=False
        super().__init__(parentItem=parentItem, **kwargs)
    
    def blockUpdates(self, flag):
        self._blockUpdates=flag
        
    def updatesBlocked(self):
        return self._blockUpdates

    @property
    def _parent(self):
        return self._GLGraphicsItem__parent

    @_parent.setter
    def _parent(self, parentItem):
        self._GLGraphicsItem__parent = parentItem

    @property
    def _children(self):
        return self._GLGraphicsItem__children

    @_children.setter
    def _children(self, childItem):
        self._GLGraphicsItem__children = childItem

    @property
    def _view(self):
        return self._GLGraphicsItem__view

    @_view.setter
    def _view(self, view):
        self._GLGraphicsItem__view = view
        
    @property
    def _visible(self):
        return self._GLGraphicsItem__visible

    @_visible.setter
    def _visible(self, flag):
        self._GLGraphicsItem__visible = flag

    @property
    def _transform(self):
        return self._GLGraphicsItem__transform

    @_transform.setter
    def _transform(self, tr):
        self._GLGraphicsItem__transform = tr

    @property
    def _initialized(self):
        return self._GLGraphicsItem__initialized

    @_initialized.setter
    def _initialized(self, flag):
        self._GLGraphicsItem__initialized=flag

    @property
    def _glOpts(self):
        return self._GLGraphicsItem__glOpts

    @_glOpts.setter
    def _glOpts(self, opts):
        self._GLGraphicsItem__glOpts = opts
        
    #@property
 #   def __flags(self):
   #     return self._GLGraphicsItem__flags
    
  #  @__flags.setter
 #   def __flags(self, flags):
      #  self._GLGraphicsItem__flags = flags

        

    def setFlag(self, flag, enabled=True, ):
        if isinstance(flag, Flag):
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsChange, enabled)
        
            if enabled:
                new_flags = self.flags() | flag
            else:
                new_flags = self.flags() & ~flag
            
            if self.flags() != new_flags:
                self._flags = new_flags
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsHaveChanged, enabled)
        
    def flags(self):    
        return self._flags
    
    def itemChange(self, change, value):
        
        if change in [self.GLGraphicsItemChange.ItemParentHasChanged, self.GLGraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()
            
            
        return value

    def setParentItem(self, item: 'GL3DGraphicsItemMixin'):
        changed_enums=[]
        if not self._parent is None:
            if item == self._parent:
                warnings.warn(f"{item.__class__.__name__} is already a parentItem to {self.__class__.__name__}, ignoring...", UserWarning)
                return
            else:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                self.itemChange(self.GLGraphicsItemChange.ItemChildRemovedChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
                self._GLGraphicsItem__parent.childItems().remove(self)
            
        if not item is None:    
            if len(changed_enums) == 0:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
            changed_enums.append((self.GLGraphicsItemChange.ItemChildAddedChange, self))
            item._GLGraphicsItem__children.append(self)
            
        if not self._view is None:
            self.itemChange(self.GLGraphicsItemChange.ItemViewChange, item)
            changed_enums.append((self.GLGraphicsItemChange.ItemViewHasChanged, item))
            self._view.removeItem(self)
        
        self._parent = item
        self._view = None
        for change, value in changed_enums:
            self.itemChange(change, value)
            
    def visible(self):
        return self._visible

    def setVisible(self, value, explicitly=None):
        self._setVisibleHelper(value, True, explicitly)

    def _setVisibleHelper(self, visible, update, explicitly):
        if explicitly:
            self._explicitlyHidden = visible 
            
        if not self.parentItem() is None and visible and not self.parentItem().visible():
            return 

        if visible == self.visible():
            return 
        newVisible = self.itemChange(self.GLGraphicsItemChange.ItemVisibleChange, visible)
        if self.visible() == newVisible:
            return
        self._visible = newVisible
        
        for child in self.childItems():
            if not visible or not child._explicitlyHidden:
                child._setVisibleHelper(visible, False, False)
        
        self.update()
        self.itemChange(self.GLGraphicsItemChange.ItemVisibleHasChanged, self.visible())

    def _setTransformHelper(self, tr):
        self.itemChange(self.GLGraphicsItemChange.ItemTransformChange, tr)
        tr = Transform3D(tr)
        if tr != self.transform():
            self._transform = tr
      #      self.itemChange(self.GLGraphicsItemChange.ItemTransformHasChanged, tr)
            self.update()

    def setTransform(self, tr):
        self._setTransformHelper(tr)
        
    def resetTransform(self):
        if self.transform().isIdentity():
            return 
        self._setTransformHelper(Transform3D())

        
    def applyTransform(self, tr, local):
        if local:
            newTransform = self.transform() * tr
        else:
            newTransform = tr * self.transform()
        self._setTransformHelper(newTransform)

    
    @abstractmethod
    def _childPaint(self):...
    
    @abstractmethod
    def _paintHelper(self):... 
    
    def paint(self):
        if not (GL3DGraphicsItemMixin.GLGraphicsItemFlag.ItemHasNoContents & GL3DGraphicsItemMixin.flags(self)): 
            self._paintHelper()
            self._childPaint()

    def update(self):
        if self.updatesBlocked():
            return
        super().update()
        
        
    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return
        self.forgetViewBox()
        
        ## check for this item's current viewbox
        view = self.getViewBox()
        
        oldView = None
        if not self._connectedView is None:
            oldView = self._connectedView()
        
        if view is oldView:
            # "  already have view", view
            return

        ## disconnect from previous view
        if not oldView is None:
            for signal, slot in [('sigRangeChanged', self.viewRangeChanged),
                                 ('sigTransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None
            
        if not view is None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        self._replaceView(oldView)        
        self.viewChanged(view, oldView)        

    def setParentWidget(self, parent=None):
        if self.parentWidget() != parent:
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetChange, parent) 
            self._parentWidget = parent
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetHasChanged, parent) 
    
    def parentWidget(self) -> GL3DViewWidget:
        return self._parentWidget

    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a ViewBox)"""
        
    def viewRect(self):
        """Return the visible bounds of this item's ViewBox or GraphicsWidget,
        in the local coordinate system of the item."""
        if self._cachedView is not None:
            return self._cachedView

        # Note that in cases of early returns here, the view cache stays empty (None).
        view = self.getViewBox()
        if view is None:
            return None
        bounds = self.mapRectFromView(view.viewRect())
        if bounds is None:
            return None

        bounds = bounds.normalized()
        
        self._cachedView = bounds
        
        ## nah.
        #for p in self.getBoundingParents():
            #bounds &= self.mapRectFromScene(p.sceneBoundingRect())

        return bounds
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if hasattr(child, "getViewBox"):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        
    def parentChanged(self):
        self.changeParent(self)
        
    def changeParent(self):
        self._updateView()

    
    @QtCore.Slot()
    def viewTransformChanged(self):
        self._cachedView = None
        
    def informViewBoundsChanged(self):
        view = self.getViewBox()
        if view is not None and hasattr(view, 'implements') and view.implements('ViewBox'):
            view.itemBoundsChanged(self)
    
    def allChildItems(self, root=None):
        """Return list of the entire item tree descending from this item."""
        if root is None:
            root = self
        tree = []
        for ch in root.childItems():
            tree.append(ch)
            tree.extend(self.allChildItems(ch))
        return tree
    
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):...

    def mapRectFromView(self, obj):
        vt = self.viewTransform()
        if vt is None:
            return None
        vt = fn.invertQTransform(vt)
        return vt.mapRect(obj)

    def getViewBox(self) -> 'GL3DViewBox':
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:  ## sometimes happens as items are being removed from a scene and collected.
                    return None
                if p is None:
                    return None
                if hasattr(type(p), 'implements'):
                    if p.implements('ViewBox'):
                        self._viewBox = weakref.ref(p)
                        break
        
        return self._viewBox() if self._viewBox else None
    
    def forgetViewBox(self):
        self._viewBox = None

    def itemTransform(self, other: 'GL3DGraphicsItemMixin', ok=None):
        if other is None:
            print("Warning: itemTransform called with null pointer")
            if not ok is None:
                ok = False
            return Transform3D()

        # Case 2: Same item
        if other == self:
            if not ok is None:
                ok = True
            return Transform3D()

        parent = self.parentItem()
        other_parent = other.parentItem()

        if parent == other:
            x = Transform3D()
            self._combine_transform_to_parent(x)
            if not ok is None:
                ok = True
            return x

        if other_parent == self:
            x = Transform3D()
            other._combine_transform_to_parent(x)
            invertible, inverted = x.inverted()
            if not ok is None:
                ok = invertible
            return inverted if invertible else Transform3D()

        # Case 5: Siblings (same parent)
        if not parent is None and parent == other_parent:
            # Fast path for simple translation
            if not self.transform().isIdentity() and not other.transform().isIdentity():
                x1 = Transform3D()
                self._combine_transform_to_parent(x1)
                
                x2 = Transform3D()
                other._combine_transform_to_parent(x2)
                
                invertible, inverted = x2.inverted()
                if not ok is None:
                    ok = invertible
                return x1 * inverted if invertible else Transform3D()
            else:
                delta = self.pos() - other.pos()
                if not ok is None:
                    ok = True
                return Transform3D.fromTranslate(delta.x(), delta.y())

        # Find closest common ancestor
        common_ancestor = self._find_common_ancestor(other)
        
        # Case 6: No common ancestor
        if common_ancestor is None:
            self.ensureSceneTransform()
            other.ensureSceneTransform()
            invertible, inverted = other.sceneTransform().inverted()
            if not ok is None:
                ok = invertible
            return self.sceneTransform() * inverted if invertible else Transform3D()

        # Case 7: Cousins (common ancestor but not direct)
        if other != common_ancestor and self != common_ancestor:
            # Transform both to common ancestor
            good = False
            this_to_common = self.itemTransform(common_ancestor, good)
            other_to_common = Transform3D()
            if good:
                other_to_common = other.itemTransform(common_ancestor, good)
            
            if not good:
                if ok is not None:
                    ok = False
                return Transform3D()
                
            invertible, inverted = other_to_common.inverted()
            if not invertible:
                if ok is not None:
                    ok = False
                return Transform3D()
                
            if ok is not None:
                ok = True
            return this_to_common * inverted

        # Case 8: Ancestor relationship
        parent_of_other = self.isAncestorOf(other)
        child = other if parent_of_other else self
        root = self if parent_of_other else other

        x = Transform3D()
        p = child
        while p is not None and p != root:
            p._combine_transform_to_parent(x)
            p = p.parentItem()

        if parent_of_other:
            invertible, inverted = x.inverted()
            if ok is not None:
                ok = invertible
            return inverted if invertible else Transform3D()
        
        if ok is not None:
            ok = True
        return x

    def _combine_transform_to_parent(self, x):
        """Helper to combine item's transform with parent transform"""
        # Apply item's transform
        t = self.transform()
        if not t.isIdentity():
            # Pre-multiply: x = t * x
            # Create a new transform since Transform3D doesn't support in-place multiplication
            x = t * x
            # Copy matrix elements to original x

        # Apply position translation
        #pos = self.pos
        #x.translate(pos.x(), pos.y())
      #  pos = 
     #   x.translate(self.pos[:,0], )

    def _find_common_ancestor(self, other):
        """Find closest common ancestor item"""
        # Collect all ancestors of this item
        ancestors = set()
        p = self
        while p is not None:
            ancestors.add(p)
            p = p.parentItem()
        
        # Find first common ancestor in other's hierarchy
        p = other
        while p is not None:
            if p in ancestors:
                return p
            p = p.parentItem()
        
        return None

    def ensureSceneTransform(self):
        """Ensure scene transform is up-to-date (dummy implementation)"""
        # In a real implementation, this would ensure the scene transform is cached
        pass

    # Required QGraphicsItem methods (simplified for example)
    def boundingRect(self):
        return self.childrenBoundingRect()




import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from PySide6 import QtWidgets, QtCore
import sys

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.gl_widget = gl.GLViewWidget()
        self.setCentralWidget(self.gl_widget)
        self.show()
        self.setWindowTitle('pyqtgraph example: GLBarGraphItem')
        self.gl_widget.setCameraPosition(distance=40)
        
        # Add grids
        self.gx = gl.GLGridItem()
        self.gx.rotate(90, 0, 1, 0)
        self.gx.translate(-10, 0, 10)
        self.gl_widget.addItem(self.gx)
        
        self.gy = gl.GLGridItem()
        self.gy.rotate(90, 1, 0, 0)
        self.gy.translate(0, -10, 10)
        self.gl_widget.addItem(self.gy)
        
        self.gz = gl.GLGridItem()
        self.gz.translate(0, 0, 0)
        self.gl_widget.addItem(self.gz)
        
    def setData(self, n):
        # Clear previous items
        for item in self.gl_widget.items[:]:
            if isinstance(item, gl.GLBarGraphItem):
                self.gl_widget.removeItem(item)
        
        # Generate data
        pos = np.empty((n, n, 3))
        pos[..., 0] = np.arange(n).reshape(n, 1)
        pos[..., 1] = np.arange(n).reshape(1, n)
        pos[..., 2] = 0
        
        size = np.empty((n, n, 3))
        size[..., 0:2] = 2*0.4
        size[..., 2] = np.random.normal(size=(n, n)) * 2 + 10
        
        # Create bar graph item
        self.bg = gl.GLBarGraphItem(pos, size)
        self.gl_widget.addItem(self.bg)

def main():
    # Check for existing QApplication before creating one
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    
  #  w1 = Window()
  #  w1.setData(10)
    
    w2 = Window()
    w2.setData(10)
    
    if sys.flags.interactive != 1 or not hasattr(QtCore, 'PYQT_VERSION'):
        sys.exit(app.exec())

if __name__ == '__main__':
    main()



import numpy as np
from dataclasses import dataclass, field
from typing import List, Dict



@dataclass(slots=True, frozen=True)
class ExpiryAttributes:
    expiry: float
    strikes: np.ndarray[float]
    n_strikes: int
    expiry_idx: int
    strike_idx_map: Dict[float, int]


def get_closest_n_strikes(center_strike, strike_arr, n):
    differences = np.abs(center_strike - strike_arr)
    sorted_indexes = np.argsort(differences)
    closest_indices = sorted_indexes[:n]
    closest_strikes = strike_arr[closest_indices]
    closest_strikes.sort()
    return closest_strikes, closest_indices, sorted_indexes[n:]



from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    
    from ...data_classes.classes import VolatilityData, VolVector
    
from PySide6 import QtWidgets
from pyvol_terminal.settings import utils as settings_utils
from . import main_view
import pyqtgraph as pg
from . import utils
from pyvol_terminal import misc_widgets
from .. import abstract_classes, custom_widgets
from . import settings, main_view
from ...quantities import engines
from ...instruments.utils import InstrumentManager
from ...data_classes import builders as builders_data_classes


class Interface(abstract_classes.ABCInterface):
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 surface_calibration_config: Dict[str, Dict[str, Any]]={},
                 tick_engine_manager=None,
                 vol_vect_container=None,
                 **configs
                 ):        
        self.instrument_manager=instrument_manager
        self.tick_engine_manager=tick_engine_manager
        settings_slots = {"Change Dimensions" : lambda:None,
                          "Toggle Price Types" : lambda:None
                          }
                          
        if tick_engine_manager is None:
            self.tick_engine_manager = engines.TickEngineManager("Strike",
                                                                 "Date",
                                                                 "Implied Volatility"
                                                                 )
        else:
            self.tick_engine_manager=tick_engine_manager
            
        volatility_data_container = builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container)


        self.widget_settings = settings.Settings(settings_slots)
        
        self._initLayout(instrument_manager,
                         surface_calibration_config["engine_container"],
                         self.tick_engine_manager,
                         self.instrument_manager.spot_instrument_container.get_objects(),
                         volatility_data_container
                         )
        
        
        
        self.all_price_types = self.instrument_manager.config["options"]["price_types"]
        self.splot_flag=True
        self.line_objects = {}
        self.displayed_price_types=["mid"]
        self.surface_flag=True
        self.scatter_flag=True
        self.subplots_flag=True


        #self.axis_transform_engine = plotting_engines.MetricEngine(self.data_container.domain.base_domain)

        self.line_objects = {px_type : {} for px_type in self.all_price_types}
        self.scatter_objects = {px_type : {} for px_type in self.all_price_types}
        self.displayed_domains=[]
    
    def _initLayout(self,
                    instrument_manager,
                    engine_container,
                    tick_engine_manager,
                    spot_objects,
                    volatility_data_container,
                    **configs
                    ):


        stacked_layout = QtWidgets.QStackedLayout()
        
        for axis, slice_type in "xy", ["smile", "term"]:         
            subplot = main_view.MainView(instrument_manager,
                                         engine_container=engine_container,
                                         tick_engine_manager=tick_engine_manager,
                                         spot_objects=spot_objects,
                                         volatility_data_container=volatility_data_container,
                                         **configs
                                         )  
            stacked_layout.addWidget(subplot)
            lines = subplot.plot.listDataItems()
            
            
            """
            domain_vect = getattr(subplot.data_container.domain, f"{axis}_vect")
            self.domain_str = [self.tick_engine_manager.function(axis, val) for val in domain_vect]
            
            colourmap = utils.legend_colour_getter(self.domain_str)
            main_view.CustomPlotDataItem.pen_map = {exp : pg.mkPen(color=colourmap[exp_str]) for exp, exp_str in zip(domain_vect.domain_vec, self.domain_str)}
            main_view.CustomScatterPlotItem.pen_map = {exp : pg.mkPen(color=colourmap[exp_str]) for exp, exp_str in zip(domain_vect.domain_vec, self.domain_str)}
            """
        
        stacked_layout.setCurrentIndex(0)
        
        central_layout = QtWidgets.QVBoxLayout()
        central_layout.addWidget(self.widget_settings)        
        central_layout.addLayout(stacked_layout)
        self.setLayout(central_layout)
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
            

    def create_plot_objects(self, px_type, value):
        self.data_container.add_slice(px_type, value)
        slice = self.data_container.slice_container[px_type][value]
        scatter = main_view.CustomPlotDataItem(x=slice.x, y=slice.y)
        line = main_view.CustomPlotDataItem(x=slice.xi, y=slice.yi)
        self.scatter_objects[px_type][value] = scatter
        self.line_objects[px_type][value] = line
        self.plot_view.add_curve(line)
        
        self.plot_view.addItem(scatter)
        line.show()
        
    def add_line(self, value):
        self.displayed_domains.append(value)
        for px_type in self.displayed_price_types:
            self.main_view.addPlot(px_type, value)
        #value_str = self.generator_func(value)
        #self.legend.add_legend_item(value_str)
        
    def remove_line(self, value):
        self.displayed_domains.remove(value)
        self.data_container.remove_slice(value)
        for px_type in self.displayed_price_types:
            self.plot_view.removeItem(self.line_objects[px_type][value])
            self.plot_view.removeItem(self.scatter_objects[px_type][value])
        value_str = self.generator_func(value)
        self.legend.remove_legend_item(value_str)
        
    def toggle_price_type222(self, px_type):
        if not px_type in self.displayed_price_types:
            self.displayed_price_types.append(px_type)
            for value in self.displayed_domains:
                self.create_plot_objects(px_type, value)
        else:
            self.displayed_price_types.remove(px_type)
            for value in self.displayed_domains:
                self.plot_view.removeItem(self.line_objects[px_type][value])
                self.plot_view.removeItem(self.scatter_objects[px_type][value])
        self.update_line_plots()
    
    def interface_switch(self):
        for instrument_object in self.instrument_manager.options_instrument_container.objects.values():
            self.data_container.update_value_by_instrument_object(instrument_object)
        self.data_container.cleanup()
        self._internal_update_interface()

    def update_line_plots(self):
        for px_type in self.displayed_price_types:
            for value in self.displayed_domains:
                slice = self.data_container.slice_container[px_type][value]
                if slice.valid_data:
                    self.scatter_objects[px_type][value].setData(x=slice.x, y=slice.y)
                    self.line_objects[px_type][value].setData(x=slice.xi, y=slice.yi)
                    
    def switch_axis(self, axis_direction, new_axis_name):
        self.tick_engine_manager.change_function(axis_direction, new_axis_name)
        
        self.data_container.switch_axis(new_axis_name, axis_direction)
        self.tick_engine.update_tick_func(new_axis_name, axis_direction)
        getattr(self, f"{axis_direction}_axis_item").setTitle(new_axis_name)
        self._internal_update_interface()
        
    def _internal_update_interface(self):
        self.update_line_plots()
        
    def update_price_cleanup(self):
        return

        
